import abc
import hashlib
from functools import cached_property
from itertools import chain, product

import numpy

import gem
from gem.utils import safe_repr


class AbstractPointSet(abc.ABC):
    """A way of specifying a known set of points, perhaps with some
    (tensor) structure.

    Points, when stored, have shape point_set_shape + (point_dimension,)
    where point_set_shape is () for scalar, (N,) for N element vector,
    (N, M) for N x M matrix etc.
    """
    def __hash__(self):
        return int.from_bytes(hashlib.md5(repr(self).encode()).digest(), byteorder="big")

    @abc.abstractmethod
    def __repr__(self):
        pass

    @property
    @abc.abstractmethod
    def points(self):
        """A flattened numpy array of points or ``UnknownPointsArray``
        object with shape (# of points, point dimension)."""

    @property
    def dimension(self):
        """Point dimension."""
        return self.points.shape[-1]

    @property
    @abc.abstractmethod
    def indices(self):
        """GEM indices with matching shape and extent to the structure of the
        point set."""

    @property
    @abc.abstractmethod
    def expression(self):
        """GEM expression describing the points, with free indices
        ``self.indices`` and shape (point dimension,)."""


class PointSingleton(AbstractPointSet):
    """A point set representing a single point.

    These have a ``gem.Literal`` expression and no indices."""

    def __init__(self, point):
        """Build a PointSingleton from a single point.

        :arg point: A single point of shape (D,) where D is the dimension of
            the point."""
        point = numpy.asarray(point)
        # 1 point ought to be a 1D array - see docstring above and points method
        assert len(point.shape) == 1
        self.point = point

    def __repr__(self):
        return f"{type(self).__name__}({safe_repr(self.point)})"

    @cached_property
    def points(self):
        # Make sure we conform to the expected (# of points, point dimension)
        # shape
        return self.point.reshape(1, -1)

    indices = ()

    @cached_property
    def expression(self):
        return gem.Literal(self.point)


class UnknownPointsArray():
    """A placeholder for a set of unknown points with appropriate length
    and size but without indexable values. For use with
    :class:`AbstractPointSet`s whose points are not known at compile
    time."""
    def __init__(self, shape):
        """
        :arg shape: The shape of the unknown set of N points of shape
            (N, D) where D is the dimension of each point.
        """
        assert len(shape) == 2
        self.shape = shape

    def __len__(self):
        return self.shape[0]


class UnknownPointSet(AbstractPointSet):
    """A point set representing a vector of points with unknown
    locations but known ``gem.Variable`` expression.

    The ``.points`` property is an `UnknownPointsArray` object with
    shape (N, D) where N is the number of points and D is their
    dimension.

    The ``.expression`` property is a derived `gem.partial_indexed` with
    shape (D,) and free indices for the points N."""

    def __init__(self, points_expr):
        r"""Build a PointSingleton from a gem expression for a single point.

        :arg points_expr: A ``gem.Variable`` expression representing a
            vector of N points in D dimensions. Should have shape (N, D)
            and no free indices. For runtime tabulation the variable
            name should begin with \'rt_:\'."""
        assert isinstance(points_expr, gem.Variable)
        assert points_expr.free_indices == ()
        assert len(points_expr.shape) == 2
        self._points_expr = points_expr

    def __repr__(self):
        return f"{type(self).__name__}({self._points_expr!r})"

    @cached_property
    def points(self):
        return UnknownPointsArray(self._points_expr.shape)

    @cached_property
    def indices(self):
        return tuple(gem.Index(extent=N) for N in self._points_expr.shape[:-1])

    @cached_property
    def expression(self):
        return gem.partial_indexed(self._points_expr, self.indices)


class PointSet(AbstractPointSet):
    """A basic point set with no internal structure representing a vector of
    points."""

    def __init__(self, points):
        """Build a PointSet from a vector of points

        :arg points: A vector of N points of shape (N, D) where D is the
            dimension of each point."""
        points = numpy.asarray(points)
        assert len(points.shape) == 2
        self.points = points

    def __repr__(self):
        return f"{type(self).__name__}({self.points!r})"

    @cached_property
    def points(self):
        pass  # set at initialisation

    @cached_property
    def indices(self):
        return tuple(gem.Index(extent=N) for N in self.points.shape[:-1])

    @cached_property
    def expression(self):
        return gem.partial_indexed(gem.Literal(self.points), self.indices)

    def almost_equal(self, other, tolerance=1e-12):
        """Approximate numerical equality of point sets"""
        return type(self) is type(other) and \
            self.points.shape == other.points.shape and \
            numpy.allclose(self.points, other.points, rtol=0, atol=tolerance)


class GaussLegendrePointSet(PointSet):
    """Gauss-Legendre quadrature points on the interval.

    This facilitates implementing discontinuous spectral elements.
    """
    def __init__(self, points):
        super().__init__(points)
        assert self.points.shape[1] == 1


class GaussLobattoLegendrePointSet(PointSet):
    """Gauss-Lobatto-Legendre quadrature points on the interval.

    This facilitates implementing continuous spectral elements.
    """
    def __init__(self, points):
        super().__init__(points)
        assert self.points.shape[1] == 1


class TensorPointSet(AbstractPointSet):

    def __init__(self, factors):
        self.factors = tuple(factors)

    def __repr__(self):
        return f"{type(self).__name__}({self.factors!r})"

    @cached_property
    def points(self):
        return numpy.array([list(chain(*pt_tuple))
                            for pt_tuple in product(*[ps.points
                                                      for ps in self.factors])])

    @cached_property
    def indices(self):
        return tuple(chain(*[ps.indices for ps in self.factors]))

    @cached_property
    def expression(self):
        result = []
        for point_set in self.factors:
            for i in range(point_set.dimension):
                result.append(gem.Indexed(point_set.expression, (i,)))
        return gem.ListTensor(result)

    def almost_equal(self, other, tolerance=1e-12):
        """Approximate numerical equality of point sets"""
        return type(self) is type(other) and \
            len(self.factors) == len(other.factors) and \
            all(s.almost_equal(o, tolerance=tolerance)
                for s, o in zip(self.factors, other.factors))


class FacetPointSet(AbstractPointSet):
    """A point set on facets.

    A FacetPointSet is constructed by mapping a lower-dimensional PointSet
    onto all facets of a higher-dimensional cell.

    Parameters
    ----------
    cell : FIAT.Cell
        The cell.
    ps : PointSet
        A lower-dimensional point set.

    """
    def __init__(self, cell, ps):
        self.cell = cell
        self.ps = ps

    def __repr__(self):
        return f"{type(self).__name__}({self.ps!r})"

    @cached_property
    def entities(self):
        """The list of all cell entites matching the reference point set dimension."""
        to_int = lambda x: sum(x) if isinstance(x, tuple) else x
        top = self.cell.topology
        return [(dim, entity)
                for dim in sorted(top)
                for entity in sorted(top[dim])
                if to_int(dim) == self.ps.dimension]

    @cached_property
    def points(self):
        """The array with the reference points mapped onto each facet."""
        ref_pts = self.ps.points
        pts = [self.cell.get_entity_transform(dim, entity)(ref_pts)
               for dim, entity in self.entities]
        return numpy.concatenate(pts)

    @cached_property
    def indices(self):
        """An Index tuple of the facet index and the reference point indices."""
        return (gem.Index(extent=len(self.entities)), *self.ps.indices)

    @cached_property
    def expression(self):
        raise NotImplementedError("Symbolic point expression is not yet implemented for FacetPointSet.")

    def almost_equal(self, other, tolerance=1e-12):
        """Approximate numerical equality of point sets"""
        return type(self) is type(other) and \
            self.cell == other.cell and \
            self.ps.almost_equal(other.ps, tolerance=tolerance)



"""Element.

This module provides an extensive list of predefined finite element
families. Users or, more likely, form compilers, may register new
elements by calling the function register_element.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Marie E. Rognes <meg@simula.no>, 2010
# Modified by Lizao Li <lzlarryli@gmail.com>, 2015, 2016
# Modified by Massimiliano Leoni, 2016
# Modified by Robert Kloefkorn, 2022
# Modified by Matthew Scroggs, 2023
# Modified by Pablo Brubeck, 2024

import warnings

from numpy import asarray

from ufl.cell import Cell, TensorProductCell
from ufl.sobolevspace import H1, H2, L2, HCurl, HDiv, HDivDiv, HCurlDiv, HEin, HInf
from ufl.utils.formatting import istr

# List of valid elements
ufl_elements = {}

# Aliases: aliases[name] (...) -> (standard_name, ...)
aliases = {}


# Function for registering new elements
def register_element(family, short_name, value_rank, sobolev_space, mapping,
                     degree_range, cellnames):
    """Register new finite element family."""
    if family in ufl_elements:
        raise ValueError(f"Finite element '{family}%s' has already been registered.")
    ufl_elements[family] = (family, short_name, value_rank, sobolev_space,
                            mapping, degree_range, cellnames)
    if short_name is not None:
        ufl_elements[short_name] = (family, short_name, value_rank, sobolev_space,
                                    mapping, degree_range, cellnames)


def register_alias(alias, to):
    """Doc."""
    aliases[alias] = to


def show_elements():
    """Shows all registered elements."""
    print("Showing all registered elements:")
    print("================================")
    shown = set()
    for k in sorted(ufl_elements.keys()):
        data = ufl_elements[k]
        if data in shown:
            continue
        shown.add(data)
        (family, short_name, value_rank, sobolev_space, mapping, degree_range, cellnames) = data
        print(f"Finite element family: '{family}', '{short_name}'")
        print(f"Sobolev space: {sobolev_space}%s")
        print(f"Mapping: {mapping}")
        print(f"Degree range: {degree_range}")
        print(f"Value rank: {value_rank}")
        print(f"Defined on cellnames: {cellnames}")
        print()


# FIXME: Consider cleanup of element names. Use notation from periodic
# table as the main, keep old names as compatibility aliases.

# NOTE: Any element with polynomial degree 0 will be considered L2,
# independent of the space passed to register_element.

# Cell groups
simplices = ("interval", "triangle", "tetrahedron", "pentatope")
cubes = ("interval", "quadrilateral", "hexahedron", "tesseract")
any_cell = (None, "vertex", *simplices, *cubes[1:], "prism", "pyramid")

# Elements in the periodic table # TODO: Register these as aliases of
# periodic table element description instead of the other way around
register_element("Lagrange", "CG", 0, H1, "identity", (1, None), any_cell)  # "P"
register_element("Brezzi-Douglas-Marini", "BDM", 1, HDiv, "contravariant Piola", (1, None), simplices[1:])  # "BDMF" (2d), "N2F" (3d)
register_element("Discontinuous Lagrange", "DG", 0, L2, "identity", (0, None), any_cell)  # "DP"
register_element("Discontinuous Taylor", "TDG", 0, L2, "identity", (0, None), simplices)
register_element("Nedelec 1st kind H(curl)", "N1curl", 1, HCurl, "covariant Piola", (1, None), simplices[1:])  # "RTE"  (2d), "N1E" (3d)
register_element("Nedelec 2nd kind H(curl)", "N2curl", 1, HCurl, "covariant Piola", (1, None), simplices[1:])  # "BDME" (2d), "N2E" (3d)
register_element("Raviart-Thomas", "RT", 1, HDiv, "contravariant Piola", (1, None), simplices[1:])   # "RTF"  (2d), "N1F" (3d)

# Elements not in the periodic table
# TODO: Implement generic Tear operator for elements instead of this:
register_element("Brezzi-Douglas-Fortin-Marini", "BDFM", 1, HDiv, "contravariant Piola", (1, None), simplices[1:])
register_element("Crouzeix-Raviart", "CR", 0, L2, "identity", (1, None), simplices[1:])
register_element("Discontinuous Raviart-Thomas", "DRT", 1, L2, "contravariant Piola", (1, None), simplices[1:])
register_element("Kong-Mulder-Veldhuizen", "KMV", 0, H1, "identity", (1, None), simplices[1:])

# Tensor elements
register_element("Regge", "Regge", 2, HEin, "double covariant Piola", (0, None), simplices)
register_element("Hellan-Herrmann-Johnson", "HHJ", 2, HDivDiv, "double contravariant Piola", (0, None), ("triangle", "tetrahedron"))
register_element("Gopalakrishnan-Lederer-Schoberl 1st kind", "GLS", 2, HCurlDiv, "covariant contravariant Piola", (1, None), simplices[1:])
register_element("Gopalakrishnan-Lederer-Schoberl 2nd kind", "GLS2", 2, HCurlDiv, "covariant contravariant Piola", (0, None), simplices[1:])

register_element("Nonconforming Arnold-Winther", "AWnc", 2, HDiv, "double contravariant Piola", (2, 2), ("triangle",))
register_element("Conforming Arnold-Winther", "AWc", 2, HDiv, "double contravariant Piola", (3, None), ("triangle",))
register_element("Hu-Zhang", "HZ", 2, HDiv, "double contravariant Piola", (3, None), ("triangle"))

# Zany elements
register_element("Bernardi-Raugel", "BR", 1, H1, "contravariant Piola", (1, None), simplices[1:])
register_element("Bernardi-Raugel Bubble", "BRB", 1, H1, "contravariant Piola", (None, None), simplices[1:])
register_element("Mardal-Tai-Winther", "MTW", 1, H1, "contravariant Piola", (3, 3), ("triangle",))
register_element("Hermite", "HER", 0, H1, "custom", (3, 3), simplices)
register_element("Argyris", "ARG", 0, H2, "custom", (5, None), ("triangle",))
register_element("Bell", "BELL", 0, H2, "custom", (5, 5), ("triangle",))
register_element("Morley", "MOR", 0, H2, "custom", (2, 2), simplices[1:])

# Macro elements
register_element("QuadraticPowellSabin6", "PS6", 0, H2, "custom", (2, 2), ("triangle",))
register_element("QuadraticPowellSabin12", "PS12", 0, H2, "custom", (2, 2), ("triangle",))
register_element("Hsieh-Clough-Tocher", "HCT", 0, H2, "custom", (3, None), ("triangle",))
register_element("Reduced-Hsieh-Clough-Tocher", "HCT-red", 0, H2, "custom", (3, 3), ("triangle",))
register_element("Johnson-Mercier", "JM", 2, HDiv, "double contravariant Piola", (1, 1), simplices[1:])

register_element("Arnold-Qin", "AQ", 1, H1, "identity", (2, 2), ("triangle",))
register_element("Reduced-Arnold-Qin", "AQ-red", 1, H1, "contravariant Piola", (2, 2), ("triangle",))
register_element("Christiansen-Hu", "CH", 1, H1, "contravariant Piola", (1, 1), simplices[1:])
register_element("Alfeld-Sorokina", "AS", 1, H1, "contravariant Piola", (2, 2), simplices[1:])

register_element("Guzman-Neilan 1st kind H1", "GN", 1, H1, "contravariant Piola", (1, None), simplices[1:])
register_element("Guzman-Neilan 2nd kind H1", "GN2", 1, H1, "contravariant Piola", (1, None), simplices[1:])
register_element("Guzman-Neilan H1(div)", "GNH1div", 1, H1, "contravariant Piola", (2, None), simplices[1:])
register_element("Guzman-Neilan Bubble", "GNB", 1, H1, "contravariant Piola", (None, None), simplices[1:])

# Special elements
register_element("Boundary Quadrature", "BQ", 0, L2, "identity", (0, None), any_cell)
register_element("Bubble", "B", 0, H1, "identity", (2, None), simplices)
register_element("FacetBubble", "FB", 0, H1, "identity", (2, None), simplices)
register_element("Quadrature", "Quadrature", 0, L2, "identity", (0, None), any_cell)
register_element("Real", "R", 0, HInf, "identity", (0, 0), any_cell + ("TensorProductCell",))
register_element("Undefined", "U", 0, L2, "identity", (0, None), any_cell)
register_element("Radau", "Rad", 0, L2, "identity", (0, None), ("interval",))
register_element("HDiv Trace", "HDivT", 0, L2, "identity", (0, None), any_cell)

# Spectral elements.
register_element("Gauss-Legendre", "GL", 0, L2, "identity", (0, None), ("interval",))
register_element("Gauss-Lobatto-Legendre", "GLL", 0, H1, "identity", (1, None), ("interval",))
register_alias("Lobatto",
               lambda family, dim, order, degree: ("Gauss-Lobatto-Legendre", order))
register_alias("Lob",
               lambda family, dim, order, degree: ("Gauss-Lobatto-Legendre", order))

register_element("Bernstein", None, 0, H1, "identity", (1, None), simplices)


# Let Nedelec H(div) elements be aliases to BDMs/RTs
register_alias("Nedelec 1st kind H(div)",
               lambda family, dim, order, degree: ("Raviart-Thomas", order))
register_alias("N1div",
               lambda family, dim, order, degree: ("Raviart-Thomas", order))

register_alias("Nedelec 2nd kind H(div)",
               lambda family, dim, order, degree: ("Brezzi-Douglas-Marini",
                                                   order))
register_alias("N2div",
               lambda family, dim, order, degree: ("Brezzi-Douglas-Marini",
                                                   order))

# Let Discontinuous Lagrange Trace element be alias to HDiv Trace
register_alias("Discontinuous Lagrange Trace",
               lambda family, dim, order, degree: ("HDiv Trace", order))
register_alias("DGT",
               lambda family, dim, order, degree: ("HDiv Trace", order))

# New elements introduced for the periodic table 2014
register_element("Q", None, 0, H1, "identity", (1, None), cubes)
register_element("DQ", None, 0, L2, "identity", (0, None), cubes)
register_element("RTCE", None, 1, HCurl, "covariant Piola", (1, None), ("quadrilateral",))
register_element("RTCF", None, 1, HDiv, "contravariant Piola", (1, None), ("quadrilateral",))
register_element("NCE", None, 1, HCurl, "covariant Piola", (1, None), ("hexahedron",))
register_element("NCF", None, 1, HDiv, "contravariant Piola", (1, None), ("hexahedron",))

register_element("S", None, 0, H1, "identity", (1, None), cubes)
register_element("DPC", None, 0, L2, "identity", (0, None), cubes)
register_element("Brezzi-Douglas-Marini Cube Edge", "BDMCE", 1, HCurl, "covariant Piola", (1, None), ("quadrilateral",))
register_element("Brezzi-Douglas-Marini Cube Face", "BDMCF", 1, HDiv, "contravariant Piola", (1, None), ("quadrilateral",))
register_element("SminusE", "SminusE", 1, HCurl, "covariant Piola", (1, None), cubes[1:3])
register_element("SminusF", "SminusF", 1, HDiv, "contravariant Piola", (1, None), cubes[1:2])
register_element("SminusDiv", "SminusDiv", 1, HDiv, "contravariant Piola", (1, None), cubes[1:3])
register_element("SminusCurl", "SminusCurl", 1, HCurl, "covariant Piola", (1, None), cubes[1:3])
register_element("AAE", None, 1, HCurl, "covariant Piola", (1, None), ("hexahedron",))
register_element("AAF", None, 1, HDiv, "contravariant Piola", (1, None), ("hexahedron",))

# New aliases introduced for the periodic table 2014
register_alias("P", lambda family, dim, order, degree: ("Lagrange", order))
register_alias("DP", lambda family, dim, order,
               degree: ("Discontinuous Lagrange", order))
register_alias("RTE", lambda family, dim, order,
               degree: ("Nedelec 1st kind H(curl)", order))
register_alias("RTF", lambda family, dim, order,
               degree: ("Raviart-Thomas", order))
register_alias("N1E", lambda family, dim, order,
               degree: ("Nedelec 1st kind H(curl)", order))
register_alias("N1F", lambda family, dim, order, degree: ("Raviart-Thomas",
                                                          order))

register_alias("BDME", lambda family, dim, order,
               degree: ("Nedelec 2nd kind H(curl)", order))
register_alias("BDMF", lambda family, dim, order,
               degree: ("Brezzi-Douglas-Marini", order))
register_alias("N2E", lambda family, dim, order,
               degree: ("Nedelec 2nd kind H(curl)", order))
register_alias("N2F", lambda family, dim, order,
               degree: ("Brezzi-Douglas-Marini", order))

# discontinuous elements using l2 pullbacks
register_element("DPC L2", None, 0, L2, "L2 Piola", (1, None), cubes)
register_element("DQ L2", None, 0, L2, "L2 Piola", (0, None), cubes)
register_element("Gauss-Legendre L2", "GL L2", 0, L2, "L2 Piola", (0, None),
                 ("interval",))
register_element("Discontinuous Lagrange L2", "DG L2", 0, L2, "L2 Piola", (0, None),
                 any_cell)  # "DP"

register_alias("DP L2", lambda family, dim, order,
               degree: ("Discontinuous Lagrange L2", order))

register_alias("P- Lambda L2", lambda family, dim, order,
               degree: feec_element_l2(family, dim, order, degree))
register_alias("P Lambda L2", lambda family, dim, order,
               degree: feec_element_l2(family, dim, order, degree))
register_alias("Q- Lambda L2", lambda family, dim, order,
               degree: feec_element_l2(family, dim, order, degree))
register_alias("S Lambda L2", lambda family, dim, order,
               degree: feec_element_l2(family, dim, order, degree))

register_alias("P- L2", lambda family, dim, order,
               degree: feec_element_l2(family, dim, order, degree))
register_alias("Q- L2", lambda family, dim, order,
               degree: feec_element_l2(family, dim, order, degree))

# mimetic spectral elements - primal and dual complexs
register_element("Extended-Gauss-Legendre", "EGL", 0, H1, "identity", (2, None),
                 ("interval",))
register_element("Extended-Gauss-Legendre Edge", "EGL-Edge", 0, L2, "identity", (1, None),
                 ("interval",))
register_element("Extended-Gauss-Legendre Edge L2", "EGL-Edge L2", 0, L2, "L2 Piola", (1, None),
                 ("interval",))
register_element("Gauss-Lobatto-Legendre Edge", "GLL-Edge", 0, L2, "identity", (0, None),
                 ("interval",))
register_element("Gauss-Lobatto-Legendre Edge L2", "GLL-Edge L2", 0, L2, "L2 Piola", (0, None),
                 ("interval",))

# directly-defined serendipity elements ala Arbogast
# currently the theory is only really worked out for quads.
register_element("Direct Serendipity", "Sdirect", 0, H1, "physical", (1, None),
                 ("quadrilateral",))
register_element("Direct Serendipity Full H(div)", "Sdirect H(div)", 1, HDiv, "physical", (1, None),
                 ("quadrilateral",))
register_element("Direct Serendipity Reduced H(div)", "Sdirect H(div) red", 1, HDiv, "physical", (1, None),
                 ("quadrilateral",))


# NOTE- the edge elements for primal mimetic spectral elements are accessed by using
# variant='mse' in the appropriate places

def feec_element(family, n, r, k):
    """Finite element exterior calculus notation.

    n = topological dimension of domain
    r = polynomial order
    k = form_degree
    """
    # Note: We always map to edge elements in 2D, don't know how to
    # differentiate otherwise?

    # Mapping from (feec name, domain dimension, form degree) to
    # (family name, polynomial order)
    _feec_elements = {
        "P- Lambda": (
            (("P", r), ("DP", r - 1)),
            (("P", r), ("RTE", r), ("DP", r - 1)),
            (("P", r), ("N1E", r), ("N1F", r), ("DP", r - 1)),
        ),
        "P Lambda": (
            (("P", r), ("DP", r)),
            (("P", r), ("BDME", r), ("DP", r)),
            (("P", r), ("N2E", r), ("N2F", r), ("DP", r)),
        ),
        "Q- Lambda": (
            (("Q", r), ("DQ", r - 1)),
            (("Q", r), ("RTCE", r), ("DQ", r - 1)),
            (("Q", r), ("NCE", r), ("NCF", r), ("DQ", r - 1)),
        ),
        "S Lambda": (
            (("S", r), ("DPC", r)),
            (("S", r), ("BDMCE", r), ("DPC", r)),
            (("S", r), ("AAE", r), ("AAF", r), ("DPC", r)),
        ),
    }

    # New notation, old verbose notation (including "Lambda") might be
    # removed
    _feec_elements["P-"] = _feec_elements["P- Lambda"]
    _feec_elements["P"] = _feec_elements["P Lambda"]
    _feec_elements["Q-"] = _feec_elements["Q- Lambda"]
    _feec_elements["S"] = _feec_elements["S Lambda"]

    family, r = _feec_elements[family][n - 1][k]

    return family, r


def feec_element_l2(family, n, r, k):
    """Finite element exterior calculus notation.

    n = topological dimension of domain
    r = polynomial order
    k = form_degree
    """
    # Note: We always map to edge elements in 2D, don't know how to
    # differentiate otherwise?

    # Mapping from (feec name, domain dimension, form degree) to
    # (family name, polynomial order)
    _feec_elements = {
        "P- Lambda L2": (
            (("P", r), ("DP L2", r - 1)),
            (("P", r), ("RTE", r), ("DP L2", r - 1)),
            (("P", r), ("N1E", r), ("N1F", r), ("DP L2", r - 1)),
        ),
        "P Lambda L2": (
            (("P", r), ("DP L2", r)),
            (("P", r), ("BDME", r), ("DP L2", r)),
            (("P", r), ("N2E", r), ("N2F", r), ("DP L2", r)),
        ),
        "Q- Lambda L2": (
            (("Q", r), ("DQ L2", r - 1)),
            (("Q", r), ("RTCE", r), ("DQ L2", r - 1)),
            (("Q", r), ("NCE", r), ("NCF", r), ("DQ L2", r - 1)),
        ),
        "S Lambda L2": (
            (("S", r), ("DPC L2", r)),
            (("S", r), ("BDMCE", r), ("DPC L2", r)),
            (("S", r), ("AAE", r), ("AAF", r), ("DPC L2", r)),
        ),
    }

    # New notation, old verbose notation (including "Lambda") might be
    # removed
    _feec_elements["P- L2"] = _feec_elements["P- Lambda L2"]
    _feec_elements["P L2"] = _feec_elements["P Lambda L2"]
    _feec_elements["Q- L2"] = _feec_elements["Q- Lambda L2"]
    _feec_elements["S L2"] = _feec_elements["S Lambda L2"]

    family, r = _feec_elements[family][n - 1][k]

    return family, r


# General FEEC notation, old verbose (can be removed)
register_alias("P- Lambda", lambda family, dim, order,
               degree: feec_element(family, dim, order, degree))
register_alias("P Lambda", lambda family, dim, order,
               degree: feec_element(family, dim, order, degree))
register_alias("Q- Lambda", lambda family, dim, order,
               degree: feec_element(family, dim, order, degree))
register_alias("S Lambda", lambda family, dim, order,
               degree: feec_element(family, dim, order, degree))

# General FEEC notation, new compact notation
register_alias("P-", lambda family, dim, order,
               degree: feec_element(family, dim, order, degree))
register_alias("Q-", lambda family, dim, order,
               degree: feec_element(family, dim, order, degree))


def canonical_element_description(family, cell, order, form_degree):
    """Given basic element information, return corresponding element information on canonical form.

    Input: family, cell, (polynomial) order, form_degree
    Output: family (canonical), short_name (for printing), order, value shape,
    reference value shape, sobolev_space.

    This is used by the FiniteElement constructor to ved input
    data against the element list and aliases defined in ufl.
    """
    # Get domain dimensions
    if cell is not None:
        tdim = cell.topological_dimension()
        if isinstance(cell, Cell):
            cellname = cell.cellname()
        else:
            cellname = None
    else:
        tdim = None
        cellname = None

    # Catch general FEEC notation "P" and "S"
    if form_degree is not None and family in ("P", "S"):
        family, order = feec_element(family, tdim, order, form_degree)

    if form_degree is not None and family in ("P L2", "S L2"):
        family, order = feec_element_l2(family, tdim, order, form_degree)

    # Check whether this family is an alias for something else
    while family in aliases:
        if tdim is None:
            raise ValueError("Need dimension to handle element aliases.")
        (family, order) = aliases[family](family, tdim, order, form_degree)

    # Check that the element family exists
    if family not in ufl_elements:
        raise ValueError(f"Unknown finite element '{family}'.")

    # Check that element data is valid (and also get common family
    # name)
    (family, short_name, value_rank, sobolev_space, mapping, krange, cellnames) = ufl_elements[family]

    # Accept CG/DG on all kind of cells, but use Q/DQ on "product" cells
    if cellname in set(cubes) - set(simplices) or isinstance(cell, TensorProductCell):
        if family == "Lagrange":
            family = "Q"
        elif family == "Discontinuous Lagrange":
            if order >= 1:
                warnings.warn("Discontinuous Lagrange element requested on %s, creating DQ element." % cell.cellname())
            family = "DQ"
        elif family == "Discontinuous Lagrange L2":
            if order >= 1:
                warnings.warn(f"Discontinuous Lagrange L2 element requested on {cell.cellname()}, "
                              "creating DQ L2 element.")
            family = "DQ L2"

    # Validate cellname if a valid cell is specified
    if not (cellname is None or cellname in cellnames):
        raise ValueError(f"Cellname '{cellname}' invalid for '{family}' finite element.")

    # Validate order if specified
    if order is not None:
        if krange is None:
            raise ValueError(f"Order {order} invalid for '{family}' finite element, should be None.")
        kmin, kmax = krange
        if not (kmin is None or (asarray(order) >= kmin).all()):
            raise ValueError(f"Order {order} invalid for '{family}' finite element.")
        if not (kmax is None or (asarray(order) <= kmax).all()):
            raise ValueError(f"Order {istr(order)} invalid for '{family}' finite element.")

    if value_rank == 2:
        # Tensor valued fundamental elements in HEin have this shape
        if tdim is None:
            raise ValueError("Cannot infer shape of element without topological and geometric dimensions.")
        reference_value_shape = (tdim, tdim)
    elif value_rank == 1:
        # Vector valued fundamental elements in HDiv and HCurl have a shape
        if tdim is None:
            raise ValueError("Cannot infer shape of element without topological and geometric dimensions.")
        reference_value_shape = (tdim,)
    elif value_rank == 0:
        # All other elements are scalar values
        reference_value_shape = ()
    else:
        raise ValueError(f"Invalid value rank {value_rank}.")

    embedded_degree = order
    if any(bubble in family for bubble in ("Guzman-Neilan", "Bernardi-Raugel")):
        embedded_degree = tdim
    return family, short_name, order, reference_value_shape, sobolev_space, mapping, embedded_degree



# -*- coding: utf-8 -*-
"""Implementation of the generalized Regge finite elements."""

# Copyright (C) 2015-2018 Lizao Li
#
# Modified by Pablo D. Brubeck (brubeck@protonmail.com), 2024
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
from FIAT import dual_set, finite_element, polynomial_set
from FIAT.check_format_variant import check_format_variant
from FIAT.functional import (PointwiseInnerProductEvaluation,
                             TensorBidirectionalIntegralMoment as BidirectionalMoment)
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature


class ReggeDual(dual_set.DualSet):
    def __init__(self, ref_el, degree, variant, qdegree):
        top = ref_el.get_topology()
        entity_ids = {dim: {i: [] for i in sorted(top[dim])} for dim in sorted(top)}
        nodes = []
        if variant == "point":
            # On a dim-facet, for all the edge tangents of the facet,
            # t^T u t is evaluated on a Pk lattice, where k = degree - dim + 1.
            for dim in sorted(top):
                for entity in sorted(top[dim]):
                    cur = len(nodes)
                    tangents = ref_el.compute_face_edge_tangents(dim, entity)
                    pts = ref_el.make_points(dim, entity, degree + 2)
                    nodes.extend(PointwiseInnerProductEvaluation(ref_el, t, t, pt)
                                 for pt in pts for t in tangents)
                    entity_ids[dim][entity].extend(range(cur, len(nodes)))

        elif variant == "integral":
            # On a dim-facet, for all the edge tangents of the facet,
            # t^T u t is integrated against a basis for Pk, where k = degree - dim + 1.
            for dim in sorted(top):
                k = degree - dim + 1
                if dim == 0 or k < 0:
                    continue
                facet = ref_el.construct_subelement(dim)
                Q = create_quadrature(facet, qdegree + k)
                P = polynomial_set.ONPolynomialSet(facet, k)
                phis = P.tabulate(Q.get_points())[(0,)*dim]
                for entity in sorted(top[dim]):
                    cur = len(nodes)
                    tangents = ref_el.compute_face_edge_tangents(dim, entity)
                    Q_mapped = FacetQuadratureRule(ref_el, dim, entity, Q)
                    detJ = Q_mapped.jacobian_determinant()
                    nodes.extend(BidirectionalMoment(ref_el, t, t/detJ, Q_mapped, phi)
                                 for phi in phis for t in tangents)
                    entity_ids[dim][entity].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class Regge(finite_element.CiarletElement):
    """The generalized Regge elements for symmetric-matrix-valued functions.
       REG(k) is the space of symmetric-matrix-valued polynomials of degree k
       or less with tangential-tangential continuity.
    """
    def __init__(self, ref_el, degree=0, variant=None):
        if degree < 0:
            raise ValueError(f"{type(self).__name__} only defined for degree >= 0")

        splitting, variant, qdegree = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        poly_set = polynomial_set.ONSymTensorPolynomialSet(ref_el, degree)
        dual = ReggeDual(ref_el, degree, variant, qdegree)
        formdegree = (1, 1)
        mapping = "double covariant piola"
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)



# Copyright (C) 2020 Robert C. Kirby (Baylor University)
#
# contributions by Keith Roberts (University of São Paulo)
# and Alexandre Olender (University of São Paulo)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
from FIAT import (
    finite_element,
    dual_set,
    functional,
    Bubble,
    FacetBubble,
    Lagrange,
    NodalEnrichedElement,
    RestrictedElement,
    reference_element,
)
from FIAT.quadrature_schemes import create_quadrature

TRIANGLE = reference_element.UFCTriangle()
TETRAHEDRON = reference_element.UFCTetrahedron()


def _get_entity_ids(ref_el, degree):
    """The topological association in a dictionary"""
    T = ref_el.topology
    sd = ref_el.get_spatial_dimension()
    if degree == 1:  # works for any spatial dimension.
        entity_ids = {0: dict((i, [i]) for i in range(len(T[0])))}
        for d in range(1, sd + 1):
            entity_ids[d] = dict((i, []) for i in range(len(T[d])))
    elif degree == 2:
        if sd == 2:
            entity_ids = {
                0: dict((i, [i]) for i in range(3)),
                1: dict((i, [i + 3]) for i in range(3)),
                2: {0: [6]},
            }
        elif sd == 3:
            entity_ids = {
                0: dict((i, [i]) for i in range(4)),
                1: dict((i, [i + 4]) for i in range(6)),
                2: dict((i, [i + 10]) for i in range(4)),
                3: {0: [14]},
            }
    elif degree == 3:
        if sd == 2:
            etop = [[3, 4], [6, 5], [7, 8]]
            entity_ids = {
                0: dict((i, [i]) for i in range(3)),
                1: dict((i, etop[i]) for i in range(3)),
                2: {0: [9, 10, 11]},
            }
        elif sd == 3:
            etop = [[4, 5], [7, 6], [8, 9], [11, 10], [12, 13], [14, 15]]
            ftop = [[16, 17, 18], [19, 20, 21], [22, 23, 24], [25, 26, 27]]
            entity_ids = {
                0: dict((i, [i]) for i in range(4)),
                1: dict((i, etop[i]) for i in range(6)),
                2: dict((i, ftop[i]) for i in range(4)),
                3: {0: [28, 29, 30, 31]},
            }
    elif degree == 4:
        if sd == 2:
            etop = [[6, 3, 7], [9, 4, 8], [10, 5, 11]]
            entity_ids = {
                0: dict((i, [i]) for i in range(3)),
                1: dict((i, etop[i]) for i in range(3)),
                2: {0: [i for i in range(12, 18)]},
            }
    elif degree == 5:
        if sd == 2:
            etop = [[9, 3, 4, 10], [12, 6, 5, 11], [13, 7, 8, 14]]
            entity_ids = {
                0: dict((i, [i]) for i in range(3)),
                1: dict((i, etop[i]) for i in range(3)),
                2: {0: [i for i in range(15, 30)]},
            }
    elif degree == 6:
        if sd == 2:
            etop = [[6, 12, 3, 13, 7], [9, 15, 4, 14, 8], [10, 16, 5, 17, 11]]
            entity_ids = {
                0: dict((i, [i]) for i in range(3)),
                1: dict((i, etop[i]) for i in range(3)),
                2: {0: [i for i in range(18, 39)]},
            }
    return entity_ids


def bump(T, deg):
    """Increase degree of polynomial along face/edges"""
    sd = T.get_spatial_dimension()
    if deg == 1:
        return (0, 0)
    else:
        if sd == 2:
            if deg < 5:
                return (1, 1)
            elif deg == 5 or deg == 6:
                return (2, 2)
            else:
                raise ValueError("Degree not supported")
        elif sd == 3:
            if deg < 4:
                return (1, 2)
            else:
                raise ValueError("Degree not supported")
        else:
            raise ValueError("Dimension of element is not supported")


def KongMulderVeldhuizenSpace(T, deg):
    sd = T.get_spatial_dimension()
    if deg == 1:
        return Lagrange(T, 1).poly_set
    else:
        L = Lagrange(T, deg)
        # Toss the bubble from Lagrange since it's dependent
        # on the higher-dimensional bubbles
        if sd == 2:
            inds = [
                i
                for i in range(L.space_dimension())
                if i not in L.dual.entity_ids[sd][0]
            ]
        elif sd == 3:
            not_inds = [L.dual.entity_ids[sd][0]] + [
                L.dual.entity_ids[sd - 1][f] for f in L.dual.entity_ids[sd - 1]
            ]
            not_inds = [item for sublist in not_inds for item in sublist]
            inds = [i for i in range(L.space_dimension()) if i not in not_inds]
        RL = RestrictedElement(L, inds)
        # interior cell bubble
        bubs = Bubble(T, deg + bump(T, deg)[1])
        if sd == 2:
            return NodalEnrichedElement(RL, bubs).poly_set
        elif sd == 3:
            # bubble on the facet
            fbubs = FacetBubble(T, deg + bump(T, deg)[0])
            return NodalEnrichedElement(RL, bubs, fbubs).poly_set


class KongMulderVeldhuizenDualSet(dual_set.DualSet):
    """The dual basis for KMV simplical elements."""

    def __init__(self, ref_el, degree):
        entity_ids = {}
        entity_ids = _get_entity_ids(ref_el, degree)
        lr = create_quadrature(ref_el, degree, scheme="KMV")
        nodes = [functional.PointEvaluation(ref_el, x) for x in lr.pts]
        super(KongMulderVeldhuizenDualSet, self).__init__(nodes, ref_el, entity_ids)


class KongMulderVeldhuizen(finite_element.CiarletElement):
    """The "lumped" simplical finite element (NB: requires custom quad. "KMV" points to achieve a diagonal mass matrix).

    References
    ----------

    Higher-order triangular and tetrahedral finite elements with mass
    lumping for solving the wave equation
    M. J. S. CHIN-JOE-KONG, W. A. MULDER and M. VAN VELDHUIZEN

    HIGHER-ORDER MASS-LUMPED FINITE ELEMENTS FOR THE WAVE EQUATION
    W.A. MULDER

    NEW HIGHER-ORDER MASS-LUMPED TETRAHEDRAL ELEMENTS
    S. GEEVERS, W.A. MULDER, AND J.J.W. VAN DER VEGT

    More Continuous Mass-Lumped Triangular Finite Elements
    W. A. MULDER

    """

    def __init__(self, ref_el, degree):
        if ref_el != TRIANGLE and ref_el != TETRAHEDRON:
            raise ValueError("KMV is only valid for triangles and tetrahedrals")
        if degree > 6 and ref_el == TRIANGLE:
            raise NotImplementedError("Only P < 7 for triangles are implemented.")
        if degree > 3 and ref_el == TETRAHEDRON:
            raise NotImplementedError("Only P < 4 for tetrahedrals are implemented.")
        S = KongMulderVeldhuizenSpace(ref_el, degree)

        dual = KongMulderVeldhuizenDualSet(ref_el, degree)
        formdegree = 0  # 0-form
        super(KongMulderVeldhuizen, self).__init__(
            S, dual, degree + max(bump(ref_el, degree)), formdegree
        )



# Copyright (C) 2024 Robert C. Kirby
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Robert C. Kirby (robert.c.kirby@gmail.com), 2024

from FIAT import dual_set, finite_element, macro, polynomial_set
from FIAT.functional import (IntegralMomentOfNormalDerivative, PointDerivative,
                             PointEvaluation)
from FIAT.jacobi import eval_jacobi_batch
from FIAT.quadrature_schemes import create_quadrature
from FIAT.reference_element import TRIANGLE, ufc_simplex


class QuadraticPowellSabin6DualSet(dual_set.DualSet):
    def __init__(self, ref_complex, degree=2):
        if degree != 2:
            raise ValueError("PS6 only defined for degree = 2")
        ref_el = ref_complex.get_parent()
        if ref_el.get_shape() != TRIANGLE:
            raise ValueError("PS6 only defined on triangles")
        top = ref_el.get_topology()
        verts = ref_el.get_vertices()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}

        # get first order jet at each vertex
        alphas = polynomial_set.mis(sd, 1)
        nodes = []

        for v in sorted(top[0]):
            pt = verts[v]
            cur = len(nodes)
            nodes.append(PointEvaluation(ref_el, pt))
            nodes.extend(PointDerivative(ref_el, pt, alpha) for alpha in alphas)
            entity_ids[0][v].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class QuadraticPowellSabin6(finite_element.CiarletElement):
    """The PS6 macroelement is a C^1 quadratic macroelement defined
    on the 6-way Powell-Sabin split of a triangle.
    """
    def __init__(self, ref_el, degree=2):
        if degree != 2:
            raise ValueError("PS6 only defined for degree = 2")
        ref_complex = macro.PowellSabinSplit(ref_el)
        dual = QuadraticPowellSabin6DualSet(ref_complex, degree)
        poly_set = macro.CkPolynomialSet(ref_complex, degree, order=1)

        super().__init__(poly_set, dual, degree)


class QuadraticPowellSabin12DualSet(dual_set.DualSet):
    def __init__(self, ref_complex, degree=2):
        if degree != 2:
            raise ValueError("PS12 only defined for degree = 2")
        ref_el = ref_complex.get_parent()
        if ref_el.get_shape() != TRIANGLE:
            raise ValueError("PS12 only defined on triangles")
        top = ref_el.get_topology()
        verts = ref_el.get_vertices()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}

        # get first order jet at each vertex
        alphas = polynomial_set.mis(sd, 1)
        nodes = []

        for v in sorted(top[0]):
            pt = verts[v]
            cur = len(nodes)
            nodes.append(PointEvaluation(ref_el, pt))
            nodes.extend(PointDerivative(ref_el, pt, alpha) for alpha in alphas)
            entity_ids[0][v].extend(range(cur, len(nodes)))

        # integral moment of normal derivatives
        rline = macro.AlfeldSplit(ufc_simplex(1))
        Q = create_quadrature(rline, degree-1)
        qpts = Q.get_points()

        x = 2.0*qpts - 1
        phis = eval_jacobi_batch(1, 1, 0, x)
        for e in sorted(top[1]):
            cur = len(nodes)
            nodes.extend(IntegralMomentOfNormalDerivative(ref_el, e, Q, phi) for phi in phis)
            entity_ids[1][e].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class QuadraticPowellSabin12(finite_element.CiarletElement):
    """The PS12 macroelement is a C^1 quadratic macroelement defined
    on the 12-way Powell-Sabin split of a triangle.
    """
    def __init__(self, ref_el, degree=2):
        if degree != 2:
            raise ValueError("PS12 only defined for degree = 2")
        ref_complex = macro.PowellSabin12Split(ref_el)
        dual = QuadraticPowellSabin12DualSet(ref_complex, degree)
        poly_set = macro.CkPolynomialSet(ref_complex, degree, order=1)

        super().__init__(poly_set, dual, degree)



from FIAT import finite_element, dual_set, macro, polynomial_set
from FIAT.functional import TensorBidirectionalIntegralMoment
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
import numpy


class JohnsonMercierDualSet(dual_set.DualSet):
    def __init__(self, ref_complex, degree, variant=None):
        if degree != 1:
            raise ValueError("Johnson-Mercier only defined for degree=1")
        if variant is not None:
            raise ValueError(f"Johnson-Mercier does not have the {variant} variant")
        ref_el = ref_complex.get_parent()
        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}
        nodes = []

        # Face dofs: bidirectional (nn and nt) Legendre moments
        dim = sd - 1
        R = numpy.array([[0, 1], [-1, 0]])
        ref_facet = ref_el.construct_subelement(dim)
        Qref = create_quadrature(ref_facet, 2*degree)
        P = polynomial_set.ONPolynomialSet(ref_facet, degree)
        phis = P.tabulate(Qref.get_points())[(0,) * dim]
        for f in sorted(top[dim]):
            cur = len(nodes)
            Q = FacetQuadratureRule(ref_el, dim, f, Qref)
            thats = ref_el.compute_tangents(dim, f)
            nhat = numpy.dot(R, *thats) if sd == 2 else numpy.cross(*thats)
            normal = nhat / Q.jacobian_determinant()
            nodes.extend(TensorBidirectionalIntegralMoment(ref_el, normal, comp, Q, phi)
                         for phi in phis for comp in (nhat, *thats))
            entity_ids[dim][f].extend(range(cur, len(nodes)))

        cur = len(nodes)
        # Interior dofs: moments for each independent component
        n = list(map(ref_el.compute_scaled_normal, sorted(top[sd-1])))
        Q = create_quadrature(ref_complex, 2*degree-1)
        P = polynomial_set.ONPolynomialSet(ref_el, degree-1, scale="L2 piola")
        phis = P.tabulate(Q.get_points())[(0,) * sd]
        nodes.extend(TensorBidirectionalIntegralMoment(ref_el, n[i+1], n[j+1], Q, phi)
                     for phi in phis for i in range(sd) for j in range(i, sd))

        entity_ids[sd][0].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class JohnsonMercier(finite_element.CiarletElement):
    """The Johnson-Mercier finite element."""

    def __init__(self, ref_el, degree=1, variant=None):
        ref_complex = macro.AlfeldSplit(ref_el)
        poly_set = macro.HDivSymPolynomialSet(ref_complex, degree)
        dual = JohnsonMercierDualSet(ref_complex, degree, variant=variant)
        formdegree = ref_el.get_spatial_dimension() - 1
        mapping = "double contravariant piola"
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)



import FIAT
import finat
import numpy as np
import pytest
from gem.interpreter import evaluate


def make_unisolvent_points(element, interior=False):
    degree = element.degree()
    ref_complex = element.get_reference_complex()
    top = ref_complex.get_topology()
    pts = []
    if interior:
        dim = ref_complex.get_spatial_dimension()
        for entity in top[dim]:
            pts.extend(ref_complex.make_points(dim, entity, degree+dim+1, variant="gll"))
    else:
        for dim in top:
            for entity in top[dim]:
                pts.extend(ref_complex.make_points(dim, entity, degree, variant="gll"))
    return pts


def check_zany_mapping(element, ref_to_phys, *args, **kwargs):
    phys_cell = ref_to_phys.phys_cell
    ref_cell = ref_to_phys.ref_cell
    phys_element = element(phys_cell, *args, **kwargs).fiat_equivalent
    finat_element = element(ref_cell, *args, **kwargs)

    ref_element = finat_element._element
    ref_cell = ref_element.get_reference_element()
    phys_cell = phys_element.get_reference_element()
    sd = ref_cell.get_spatial_dimension()

    shape = ref_element.value_shape()
    ref_pts = make_unisolvent_points(ref_element, interior=True)
    ref_vals = ref_element.tabulate(0, ref_pts)[(0,)*sd]

    phys_pts = make_unisolvent_points(phys_element, interior=True)
    phys_vals = phys_element.tabulate(0, phys_pts)[(0,)*sd]

    mapping = ref_element.mapping()[0]
    if mapping == "affine":
        ref_vals_piola = ref_vals
    else:
        # Piola map the reference elements
        J, b = FIAT.reference_element.make_affine_mapping(ref_cell.vertices,
                                                          phys_cell.vertices)
        K = []
        if "covariant" in mapping:
            K.append(np.linalg.inv(J).T)
        if "contravariant" in mapping:
            K.append(J / np.linalg.det(J))

        if len(shape) == 2:
            piola_map = lambda x: K[0] @ x @ K[-1].T
        else:
            piola_map = lambda x: K[0] @ x

        ref_vals_piola = np.zeros(ref_vals.shape)
        for i in range(ref_vals.shape[0]):
            for k in range(ref_vals.shape[-1]):
                ref_vals_piola[i, ..., k] = piola_map(ref_vals[i, ..., k])

    # Zany map the results
    num_bfs = phys_element.space_dimension()
    num_dofs = finat_element.space_dimension()
    try:
        Mgem = finat_element.basis_transformation(ref_to_phys)
        M = evaluate([Mgem])[0].arr
        ref_vals_zany = np.tensordot(M, ref_vals_piola, (-1, 0))
    except AttributeError:
        M = np.eye(num_dofs, num_bfs)
        ref_vals_zany = ref_vals_piola

    # Solve for the basis transformation and compare results
    Phi = ref_vals_piola.reshape(num_bfs, -1)
    phi = phys_vals.reshape(num_bfs, -1)
    Vh, residual, *_ = np.linalg.lstsq(Phi.T, phi.T)
    Mh = Vh.T
    Mh = Mh[:num_dofs]
    Mh[abs(Mh) < 1E-10] = 0
    M[abs(M) < 1E-10] = 0

    with np.errstate(divide='ignore', invalid='ignore'):
        error = M.T / Mh.T - 1
    error[error != error] = 0
    error[abs(error) < 1E-10] = 0
    error = error[np.ix_(*map(np.unique, np.nonzero(error)))]
    error[error != 0] += 1

    assert np.allclose(residual, 0), str(error)
    assert np.allclose(ref_vals_zany, phys_vals[:num_dofs]), str(error)


@pytest.mark.parametrize("element", [
                         finat.Morley,
                         finat.Hermite,
                         finat.Bell,
                         ])
def test_C1_triangle(ref_to_phys, element):
    check_zany_mapping(element, ref_to_phys[2])


@pytest.mark.parametrize("element", [
                         finat.Morley,
                         ])
def test_C1_tetrahedron(ref_to_phys, element):
    check_zany_mapping(element, ref_to_phys[3])


@pytest.mark.parametrize("element", [
                         finat.QuadraticPowellSabin6,
                         finat.QuadraticPowellSabin12,
                         finat.ReducedHsiehCloughTocher,
                         ])
def test_C1_macroelements(ref_to_phys, element):
    kwargs = {}
    if element == finat.QuadraticPowellSabin12:
        kwargs = dict(avg=True)
    check_zany_mapping(element, ref_to_phys[2], **kwargs)


@pytest.mark.parametrize("element, degree", [
    *((finat.Argyris, k) for k in range(5, 8)),
    *((finat.HsiehCloughTocher, k) for k in range(3, 6))
])
def test_high_order_C1_elements(ref_to_phys, element, degree):
    check_zany_mapping(element, ref_to_phys[2], degree, avg=True)


def test_argyris_point(ref_to_phys):
    check_zany_mapping(finat.Argyris, ref_to_phys[2], variant="point")


zany_piola_elements = {
    2: [
        finat.MardalTaiWinther,
        finat.ReducedArnoldQin,
        finat.ArnoldWinther,
        finat.ArnoldWintherNC,
    ],
    3: [
        finat.BernardiRaugel,
        finat.BernardiRaugelBubble,
        finat.AlfeldSorokina,
        finat.ChristiansenHu,
        finat.JohnsonMercier,
        finat.GuzmanNeilanFirstKindH1,
        finat.GuzmanNeilanSecondKindH1,
        finat.GuzmanNeilanBubble,
        finat.GuzmanNeilanH1div,
    ],
}


@pytest.mark.parametrize("dimension, element", [
                         *((2, e) for e in zany_piola_elements[2]),
                         *((2, e) for e in zany_piola_elements[3]),
                         *((3, e) for e in zany_piola_elements[3]),
                         ])
def test_piola(ref_to_phys, element, dimension):
    check_zany_mapping(element, ref_to_phys[dimension])


@pytest.mark.parametrize("element, degree, variant", [
    *((finat.HuZhang, k, v) for v in ("integral", "point") for k in range(3, 6)),
])
def test_piola_triangle_high_order(ref_to_phys, element, degree, variant):
    check_zany_mapping(element, ref_to_phys[2], degree, variant)


@pytest.mark.parametrize("element, degree", [
                         *((finat.Regge, k) for k in range(3)),
                         *((finat.HellanHerrmannJohnson, k) for k in range(3)),
                         *((finat.GopalakrishnanLedererSchoberlFirstKind, k) for k in range(1, 4)),
                         *((finat.GopalakrishnanLedererSchoberlSecondKind, k) for k in range(0, 3)),
                         ])
@pytest.mark.parametrize("dimension", [2, 3])
@pytest.mark.parametrize("variant", [None, "alfeld"])
def test_affine(ref_to_phys, element, degree, variant, dimension):
    check_zany_mapping(element, ref_to_phys[dimension], degree, variant=variant)


@pytest.mark.parametrize("element", [finat.BrezziDouglasMarini, finat.NedelecSecondKind])
@pytest.mark.parametrize("degree", [1, 2])
@pytest.mark.parametrize("dimension", [2, 3])
@pytest.mark.parametrize("variant", [None, "iso"])
def test_macro_piola(ref_to_phys, element, degree, variant, dimension):
    check_zany_mapping(element, ref_to_phys[dimension], degree, variant=variant)



# -*- coding: utf-8 -*-
"""Implementation of the Mardal-Tai-Winther finite elements."""

# Copyright (C) 2020 by Robert C. Kirby (Baylor University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


from FIAT import dual_set, expansions, finite_element, polynomial_set
from FIAT.functional import (IntegralMomentOfNormalEvaluation,
                             IntegralMomentOfTangentialEvaluation,
                             IntegralLegendreNormalMoment,
                             IntegralMomentOfDivergence)

from FIAT.quadrature_schemes import create_quadrature


def DivergenceDubinerMoments(ref_el, start_deg, stop_deg, comp_deg):
    sd = ref_el.get_spatial_dimension()
    P = polynomial_set.ONPolynomialSet(ref_el, stop_deg)
    Q = create_quadrature(ref_el, comp_deg + stop_deg)

    dim0 = expansions.polynomial_dimension(ref_el, start_deg-1)
    dim1 = expansions.polynomial_dimension(ref_el, stop_deg)
    indices = list(range(dim0, dim1))
    phis = P.take(indices).tabulate(Q.get_points())[(0,)*sd]
    for phi in phis:
        yield IntegralMomentOfDivergence(ref_el, Q, phi)


class MardalTaiWintherDual(dual_set.DualSet):
    """Degrees of freedom for Mardal-Tai-Winther elements."""
    def __init__(self, ref_el, degree):
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()

        if sd != 2:
            raise ValueError("Mardal-Tai-Winther elements are only defined in dimension 2.")

        if degree != 3:
            raise ValueError("Mardal-Tai-Winther elements are only defined for degree 3.")

        entity_ids = {dim: {entity: [] for entity in top[dim]} for dim in top}
        nodes = []

        # no vertex dofs

        # On each facet, let n be its normal.  We need to integrate
        # u.n and u.t against the first Legendre polynomial (constant)
        # and u.n against the second (linear).
        facet = ref_el.get_facet_element()
        # Facet nodes are \int_F v.n p ds where p \in P_{q-1}
        # degree is q - 1
        Q = create_quadrature(facet, degree+1)
        Pq = polynomial_set.ONPolynomialSet(facet, 1)
        phis = Pq.tabulate(Q.get_points())[(0,)*(sd - 1)]
        for f in sorted(top[sd-1]):
            cur = len(nodes)
            nodes.append(IntegralMomentOfNormalEvaluation(ref_el, Q, phis[0], f))
            nodes.append(IntegralMomentOfTangentialEvaluation(ref_el, Q, phis[0], f))
            nodes.append(IntegralMomentOfNormalEvaluation(ref_el, Q, phis[1], f))
            entity_ids[sd-1][f].extend(range(cur, len(nodes)))

        # Generate constraint nodes on the cell and facets
        # * div(v) must be constant on the cell.  Since v is a cubic and
        #   div(v) is quadratic, we need the integral of div(v) against the
        #   linear and quadratic Dubiner polynomials to vanish.
        #   There are two linear and three quadratics, so these are five
        #   constraints
        # * v.n must be linear on each facet.  Since v.n is cubic, we need
        #   the integral of v.n against the cubic and quadratic Legendre
        #   polynomial to vanish on each facet.

        # So we introduce functionals whose kernel describes this property,
        # as described in the FIAT paper.
        start_order = 2
        stop_order = 3
        qdegree = degree + stop_order
        for f in sorted(top[sd-1]):
            cur = len(nodes)
            nodes.extend(IntegralLegendreNormalMoment(ref_el, f, order, qdegree)
                         for order in range(start_order, stop_order+1))
            entity_ids[sd-1][f].extend(range(cur, len(nodes)))

        cur = len(nodes)
        nodes.extend(DivergenceDubinerMoments(ref_el, start_order-1, stop_order-1, degree))
        entity_ids[sd][0].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class MardalTaiWinther(finite_element.CiarletElement):
    """The definition of the Mardal-Tai-Winther element.
    """
    def __init__(self, ref_el, degree=3):
        sd = ref_el.get_spatial_dimension()
        assert degree == 3, "Only defined for degree 3"
        assert sd == 2, "Only defined for dimension 2"
        poly_set = polynomial_set.ONPolynomialSet(ref_el, degree, (sd,))
        dual = MardalTaiWintherDual(ref_el, degree)
        formdegree = sd-1
        mapping = "contravariant piola"
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)



# Copyright (C) 2010-2012 Marie E. Rognes
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numpy

from FIAT.finite_element import CiarletElement
from FIAT.dual_set import DualSet
from FIAT.polynomial_set import ONPolynomialSet
from FIAT.functional import PointEdgeTangentEvaluation as Tangent
from FIAT.functional import FrobeniusIntegralMoment as IntegralMoment
from FIAT.raviart_thomas import RaviartThomas
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
from FIAT.check_format_variant import check_format_variant


class NedelecSecondKindDual(DualSet):
    r"""
    This class represents the dual basis for the Nedelec H(curl)
    elements of the second kind. The degrees of freedom (L) for the
    elements of the k'th degree are

    d = 2:

      vertices: None

      edges:    L(f) = f (x_i) * t       for (k+1) points x_i on each edge

      cell:     L(f) = \int f * g * dx   for g in RT_{k-1}


    d = 3:

      vertices: None

      edges:    L(f)  = f(x_i) * t         for (k+1) points x_i on each edge

      faces:    L(f) = \int_F f * g * ds   for g in RT_{k-1}(F) for each face F

      cell:     L(f) = \int f * g * dx     for g in RT_{k-2}

    Higher spatial dimensions are not yet implemented. (For d = 1,
    these elements coincide with the CG_k elements.)
    """

    def __init__(self, cell, degree, variant, interpolant_deg):

        # Define degrees of freedom
        (dofs, ids) = self.generate_degrees_of_freedom(cell, degree, variant, interpolant_deg)
        # Call init of super-class
        super().__init__(dofs, cell, ids)

    def generate_degrees_of_freedom(self, cell, degree, variant, interpolant_deg):
        "Generate dofs and geometry-to-dof maps (ids)."

        dofs = []
        ids = {}

        # Extract spatial dimension and topology
        d = cell.get_spatial_dimension()
        assert (d in (2, 3)), "Second kind Nedelecs only implemented in 2/3D."

        # Zero vertex-based degrees of freedom
        ids[0] = {i: [] for i in sorted(cell.topology[0])}

        # (degree+1) degrees of freedom per entity of codimension 1 (edges)
        (edge_dofs, ids[1]) = self._generate_edge_dofs(cell, degree, 0, variant, interpolant_deg)
        dofs.extend(edge_dofs)

        # Include face degrees of freedom if 3D
        if d == 3:
            face_dofs, ids[d-1] = self._generate_facet_dofs(d-1, cell, degree,
                                                            len(dofs), variant, interpolant_deg)
            dofs.extend(face_dofs)

        # Varying degrees of freedom (possibly zero) per cell
        cell_dofs, ids[d] = self._generate_facet_dofs(d, cell, degree, len(dofs), variant, interpolant_deg)
        dofs.extend(cell_dofs)

        return (dofs, ids)

    def _generate_edge_dofs(self, cell, degree, offset, variant, interpolant_deg):
        """Generate degrees of freedom (dofs) for entities of
        codimension 1 (edges)."""

        if variant == "integral":
            return self._generate_facet_dofs(1, cell, degree, offset, variant, interpolant_deg)

        # (degree+1) tangential component point evaluation degrees of
        # freedom per entity of codimension 1 (edges)
        dofs = []
        ids = {}
        if variant == "point":
            for edge in range(len(cell.get_topology()[1])):

                # Create points for evaluation of tangential components
                points = cell.make_points(1, edge, degree + 2)

                # A tangential component evaluation for each point
                dofs.extend(Tangent(cell, edge, point) for point in points)

                # Associate these dofs with this edge
                i = len(points) * edge
                ids[edge] = list(range(offset + i, offset + i + len(points)))

        return (dofs, ids)

    def _generate_facet_dofs(self, dim, cell, degree, offset, variant, interpolant_deg):
        """Generate degrees of freedom (dofs) for facets."""

        # Initialize empty dofs and identifiers (ids)
        num_facets = len(cell.get_topology()[dim])
        dofs = []
        ids = {i: [] for i in range(num_facets)}

        # Return empty info if not applicable
        rt_degree = degree - dim + 1
        if rt_degree < 1:
            return (dofs, ids)

        if interpolant_deg is None:
            interpolant_deg = degree

        # Construct quadrature scheme for the reference facet
        ref_facet = cell.construct_subelement(dim)
        Q_ref = create_quadrature(ref_facet, interpolant_deg + rt_degree)
        if dim == 1:
            Phi = ONPolynomialSet(ref_facet, rt_degree, (dim,))
        else:
            # Construct Raviart-Thomas on the reference facet
            RT = RaviartThomas(ref_facet, rt_degree, variant)
            Phi = RT.get_nodal_basis()

        # Evaluate basis functions at reference quadrature points
        Phis = Phi.tabulate(Q_ref.get_points())[(0,) * dim]
        # Note: Phis has dimensions:
        # num_basis_functions x num_components x num_quad_points
        Phis = numpy.transpose(Phis, (0, 2, 1))
        # Note: Phis has dimensions:
        # num_basis_functions x num_quad_points x num_components

        # Iterate over the facets
        cur = offset
        for facet in range(num_facets):
            # Get the quadrature and Jacobian on this facet
            Q_facet = FacetQuadratureRule(cell, dim, facet, Q_ref)
            J = Q_facet.jacobian()
            detJ = Q_facet.jacobian_determinant()

            # Map Phis -> phis (reference values to physical values)
            piola_map = J / detJ
            phis = numpy.dot(Phis, piola_map.T)
            phis = numpy.transpose(phis, (0, 2, 1))

            # Construct degrees of freedom as integral moments on this cell,
            # using the face quadrature weighted against the values
            # of the (physical) Raviart--Thomas'es on the face
            dofs.extend(IntegralMoment(cell, Q_facet, phi) for phi in phis)

            # Assign identifiers (num RTs per face + previous edge dofs)
            ids[facet].extend(range(cur, cur + len(phis)))
            cur += len(phis)

        return (dofs, ids)


class NedelecSecondKind(CiarletElement):
    """
    The H(curl) Nedelec elements of the second kind on triangles and
    tetrahedra: the polynomial space described by the full polynomials
    of degree k, with a suitable set of degrees of freedom to ensure
    H(curl) conformity.

    :arg ref_el: The reference element.
    :arg degree: The degree.
    :arg variant: optional variant specifying the types of nodes.

    variant can be chosen from ["point", "integral", "integral(q)"]
    "point" -> dofs are evaluated by point evaluation. Note that this variant
    has suboptimal convergence order in the H(curl)-norm
    "integral" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence.
    "integral(q)" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence plus q. You might want to choose a high
    quadrature degree to make sure that expressions will be interpolated
    exactly. This is important when you want to have (nearly) curl-preserving
    interpolation.
    """

    def __init__(self, ref_el, degree, variant=None):
        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        # Check degree
        assert degree >= 1, "Second kind Nedelecs start at 1!"

        # Get dimension
        d = ref_el.get_spatial_dimension()
        # Construct polynomial basis for d-vector fields
        Ps = ONPolynomialSet(ref_el, degree, (d, ))

        # Construct dual space
        Ls = NedelecSecondKindDual(ref_el, degree, variant, interpolant_deg)

        # Set form degree
        formdegree = 1  # 1-form

        # Set mapping
        mapping = "covariant piola"

        # Call init of super-class
        super().__init__(Ps, Ls, degree, formdegree, mapping=mapping)



"""This module defines the UFL finite element classes."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Anders Logg 2014
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

import numpy as np

from ufl.cell import as_cell
from finat.ufl.finiteelement import FiniteElement
from finat.ufl.finiteelementbase import FiniteElementBase
from ufl.permutation import compute_indices
from ufl.pullback import MixedPullback, SymmetricPullback
from ufl.utils.indexflattening import flatten_multiindex, shape_to_strides, unflatten_index
from ufl.utils.sequences import max_degree, product


class MixedElement(FiniteElementBase):
    """A finite element composed of a nested hierarchy of mixed or simple elements."""
    __slots__ = ("_sub_elements", "_cells")

    def __init__(self, *elements, **kwargs):
        """Create mixed finite element from given list of elements."""
        if type(self) is MixedElement:
            if kwargs:
                raise ValueError("Not expecting keyword arguments to MixedElement constructor.")

        # Un-nest arguments if we get a single argument with a list of elements
        if len(elements) == 1 and isinstance(elements[0], (tuple, list)):
            elements = elements[0]
        # Interpret nested tuples as sub-mixedelements recursively
        elements = [MixedElement(e) if isinstance(e, (tuple, list)) else e
                    for e in elements]
        self._sub_elements = elements

        # Pick the first cell, for now all should be equal
        cells = tuple(sorted(set(element.cell for element in elements) - set([None])))
        self._cells = cells
        if cells:
            cell = cells[0]
            # Require that all elements are defined on the same cell
            if not all(c == cell for c in cells[1:]):
                raise ValueError("Sub elements must live on the same cell.")
        else:
            cell = None

        # Check that all elements use the same quadrature scheme TODO:
        # We can allow the scheme not to be defined.
        if len(elements) == 0:
            quad_scheme = None
        else:
            quad_scheme = elements[0].quadrature_scheme()
            if not all(e.quadrature_scheme() == quad_scheme for e in elements):
                raise ValueError("Quadrature scheme mismatch for sub elements of mixed element.")

        # Compute value sizes in global and reference configurations
        reference_value_size_sum = sum(product(s.reference_value_shape) for s in self._sub_elements)

        # Default reference value shape: Treated simply as all
        # subelement reference values unpacked in a vector.
        reference_value_shape = kwargs.get('reference_value_shape', (reference_value_size_sum,))

        # Initialize element data
        degrees = {e.degree() for e in self._sub_elements} - {None}
        degree = max_degree(degrees) if degrees else None
        FiniteElementBase.__init__(self, "Mixed", cell, degree, quad_scheme,
                                   reference_value_shape)

    def __repr__(self):
        """Doc."""
        return "MixedElement(" + ", ".join(repr(e) for e in self._sub_elements) + ")"

    def _is_linear(self):
        """Doc."""
        return all(i._is_linear() for i in self._sub_elements)

    def reconstruct_from_elements(self, *elements):
        """Reconstruct a mixed element from new subelements."""
        if all(a == b for (a, b) in zip(elements, self._sub_elements)):
            return self
        return MixedElement(*elements)

    def symmetry(self, domain):
        r"""Return the symmetry dict, which is a mapping :math:`c_0 \\to c_1`.

        meaning that component :math:`c_0` is represented by component
        :math:`c_1`.
        A component is a tuple of one or more ints.
        """
        # Build symmetry map from symmetries of subelements
        sm = {}
        # Base index of the current subelement into mixed value
        j = 0
        for e in self._sub_elements:
            sh = e.value_shape(domain)
            st = shape_to_strides(sh)
            # Map symmetries of subelement into index space of this
            # element
            for c0, c1 in e.symmetry().items():
                j0 = flatten_multiindex(c0, st) + j
                j1 = flatten_multiindex(c1, st) + j
                sm[(j0,)] = (j1,)
            # Update base index for next element
            j += product(sh)
        if j != product(self.value_shape(domain)):
            raise ValueError("Size mismatch in symmetry algorithm.")
        return sm or {}

    @property
    def sobolev_space(self):
        """Doc."""
        return max(e.sobolev_space for e in self._sub_elements)

    def mapping(self):
        """Doc."""
        if all(e.mapping() == "identity" for e in self._sub_elements):
            return "identity"
        else:
            return "undefined"

    @property
    def num_sub_elements(self):
        """Return number of sub elements."""
        return len(self._sub_elements)

    @property
    def sub_elements(self):
        """Return list of sub elements."""
        return self._sub_elements

    def extract_subelement_component(self, domain, i):
        """Extract direct subelement index and subelement relative.

        component index for a given component index.
        """
        if isinstance(i, int):
            i = (i,)
        self._check_component(i)

        # Select between indexing modes
        if len(self.value_shape(domain)) == 1:
            # Indexing into a long vector of flattened subelement
            # shapes
            j, = i

            # Find subelement for this index
            for sub_element_index, e in enumerate(self._sub_elements):
                sh = e.value_shape(domain)
                si = product(sh)
                if j < si:
                    break
                j -= si
            if j < 0:
                raise ValueError("Moved past last value component!")

            # Convert index into a shape tuple
            st = shape_to_strides(sh)
            component = unflatten_index(j, st)
        else:
            # Indexing into a multidimensional tensor where subelement
            # index is first axis
            sub_element_index = i[0]
            if sub_element_index >= len(self._sub_elements):
                raise ValueError(f"Illegal component index (dimension {sub_element_index}).")
            component = i[1:]
        return (sub_element_index, component)

    def extract_component(self, i):
        """Recursively extract component index relative to a (simple) element.

        and that element for given value component index.
        """
        sub_element_index, component = self.extract_subelement_component(i)
        return self._sub_elements[sub_element_index].extract_component(component)

    def extract_subelement_reference_component(self, i):
        """Extract direct subelement index and subelement relative.

        reference_component index for a given reference_component index.
        """
        if isinstance(i, int):
            i = (i,)
        self._check_reference_component(i)

        # Select between indexing modes
        assert len(self.reference_value_shape) == 1
        # Indexing into a long vector of flattened subelement shapes
        j, = i

        # Find subelement for this index
        for sub_element_index, e in enumerate(self._sub_elements):
            sh = e.reference_value_shape
            si = product(sh)
            if j < si:
                break
            j -= si
        if j < 0:
            raise ValueError("Moved past last value reference_component!")

        # Convert index into a shape tuple
        st = shape_to_strides(sh)
        reference_component = unflatten_index(j, st)
        return (sub_element_index, reference_component)

    def extract_reference_component(self, i):
        """Recursively extract reference_component index relative to a (simple) element.

        and that element for given value reference_component index.
        """
        sub_element_index, reference_component = self.extract_subelement_reference_component(i)
        return self._sub_elements[sub_element_index].extract_reference_component(reference_component)

    def is_cellwise_constant(self, component=None):
        """Return whether the basis functions of this element is spatially constant over each cell."""
        if component is None:
            return all(e.is_cellwise_constant() for e in self.sub_elements)
        else:
            i, e = self.extract_component(component)
            return e.is_cellwise_constant()

    def degree(self, component=None):
        """Return polynomial degree of finite element."""
        if component is None:
            return self._degree  # from FiniteElementBase, computed as max of subelements in __init__
        else:
            i, e = self.extract_component(component)
            return e.degree()

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        return min(e.embedded_subdegree for e in self.sub_elements)

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        return max(e.embedded_superdegree for e in self.sub_elements)

    def reconstruct(self, **kwargs):
        """Doc."""
        return MixedElement(*[e.reconstruct(**kwargs) for e in self.sub_elements])

    def variant(self):
        """Doc."""
        try:
            variant, = {e.variant() for e in self.sub_elements}
            return variant
        except ValueError:
            return None

    def __str__(self):
        """Format as string for pretty printing."""
        tmp = ", ".join(str(element) for element in self._sub_elements)
        return "<Mixed element: (" + tmp + ")>"

    def shortstr(self):
        """Format as string for pretty printing."""
        tmp = ", ".join(element.shortstr() for element in self._sub_elements)
        return "Mixed<" + tmp + ">"

    @property
    def pullback(self):
        """Get the pull back."""
        return MixedPullback(self)


class VectorElement(MixedElement):
    """A special case of a mixed finite element where all elements are equal."""

    __slots__ = ("_repr", "_mapping", "_sub_element")

    def __init__(self, family, cell=None, degree=None, dim=None,
                 form_degree=None, quad_scheme=None, variant=None):
        """Create vector element (repeated mixed element)."""
        if isinstance(family, FiniteElementBase):
            sub_element = family
            cell = sub_element.cell
            variant = sub_element.variant()
        else:
            if cell is not None:
                cell = as_cell(cell)
            # Create sub element
            sub_element = FiniteElement(family, cell, degree,
                                        form_degree=form_degree,
                                        quad_scheme=quad_scheme,
                                        variant=variant)

        # Set default size if not specified
        if dim is None:
            if cell is None:
                raise ValueError("Cannot infer vector dimension without a cell.")
            # TODO: is this the right default
            dim = cell.topological_dimension()

        self._mapping = sub_element.mapping()
        # Create list of sub elements for mixed element constructor
        sub_elements = [sub_element] * dim

        # Compute value shapes
        reference_value_shape = (dim,) + sub_element.reference_value_shape

        # Initialize element data
        MixedElement.__init__(self, sub_elements,
                              reference_value_shape=reference_value_shape)

        FiniteElementBase.__init__(self, sub_element.family(), sub_element.cell, sub_element.degree(),
                                   sub_element.quadrature_scheme(), reference_value_shape)

        self._sub_element = sub_element

        if variant is None:
            var_str = ""
        else:
            var_str = ", variant='" + variant + "'"

        # Cache repr string
        self._repr = f"VectorElement({repr(sub_element)}, dim={dim}{var_str})"

    def __repr__(self):
        """Doc."""
        return self._repr

    def reconstruct(self, **kwargs):
        """Doc."""
        sub_element = self._sub_element.reconstruct(**kwargs)
        return VectorElement(sub_element, dim=len(self.sub_elements))

    def variant(self):
        """Return the variant used to initialise the element."""
        return self._sub_element.variant()

    def mapping(self):
        """Doc."""
        return self._mapping

    def __str__(self):
        """Format as string for pretty printing."""
        return ("<vector element with %d components of %s>" %
                (len(self._sub_elements), self._sub_element))

    def shortstr(self):
        """Format as string for pretty printing."""
        return "Vector<%d x %s>" % (len(self._sub_elements),
                                    self._sub_element.shortstr())

    @property
    def pullback(self):
        """Get the pull back."""
        return self._sub_element.pullback


class TensorElement(MixedElement):
    """A special case of a mixed finite element where all elements are equal."""
    __slots__ = ("_sub_element", "_shape", "_symmetry",
                 "_sub_element_mapping",
                 "_flattened_sub_element_mapping",
                 "_mapping", "_repr")

    def __init__(self, family, cell=None, degree=None, shape=None,
                 symmetry=None, quad_scheme=None, variant=None):
        """Create tensor element (repeated mixed element with optional symmetries)."""
        if isinstance(family, FiniteElementBase):
            sub_element = family
            cell = sub_element.cell
            variant = sub_element.variant()
        else:
            if cell is not None:
                cell = as_cell(cell)
            # Create scalar sub element
            sub_element = FiniteElement(family, cell, degree, quad_scheme=quad_scheme,
                                        variant=variant)

        # Set default shape if not specified
        if shape is None:
            if cell is None:
                raise ValueError("Cannot infer tensor shape without a cell.")
            # TODO: is this the right default
            dim = cell.topological_dimension()
            shape = (dim, dim)

        if symmetry is None:
            symmetry = {}
        elif symmetry is True:
            # Construct default symmetry dict for matrix elements
            if not (len(shape) == 2 and shape[0] == shape[1]):
                raise ValueError("Cannot set automatic symmetry for non-square tensor.")
            symmetry = dict(((i, j), (j, i)) for i in range(shape[0])
                            for j in range(shape[1]) if i > j)
        else:
            if not isinstance(symmetry, dict):
                raise ValueError("Expecting symmetry to be None (unset), True, or dict.")

        # Validate indices in symmetry dict
        for i, j in symmetry.items():
            if len(i) != len(j):
                raise ValueError("Non-matching length of symmetry index tuples.")
            for k in range(len(i)):
                if not (i[k] >= 0 and j[k] >= 0 and i[k] < shape[k] and j[k] < shape[k]):
                    raise ValueError("Symmetry dimensions out of bounds.")

        # Compute all index combinations for given shape
        indices = compute_indices(shape)

        # Compute mapping from indices to sub element number,
        # accounting for symmetry
        sub_elements = []
        sub_element_mapping = {}
        for index in indices:
            if index in symmetry:
                continue
            sub_element_mapping[index] = len(sub_elements)
            sub_elements += [sub_element]

        # Update mapping for symmetry
        for index in indices:
            if index in symmetry:
                sub_element_mapping[index] = sub_element_mapping[symmetry[index]]
        flattened_sub_element_mapping = [sub_element_mapping[index] for i,
                                         index in enumerate(indices)]

        # Compute reference value shape based on symmetries
        if symmetry:
            reference_value_shape = (product(shape) - len(symmetry),)
            self._mapping = "symmetries"
        else:
            reference_value_shape = shape
            self._mapping = sub_element.mapping()

        reference_value_shape = reference_value_shape + sub_element.reference_value_shape
        # Initialize element data
        MixedElement.__init__(self, sub_elements,
                              reference_value_shape=reference_value_shape)
        self._family = sub_element.family()
        self._degree = sub_element.degree()
        self._sub_element = sub_element
        self._shape = shape
        self._symmetry = symmetry
        self._sub_element_mapping = sub_element_mapping
        self._flattened_sub_element_mapping = flattened_sub_element_mapping

        if variant is None:
            var_str = ""
        else:
            var_str = ", variant='" + variant + "'"

        # Cache repr string
        self._repr = (f"TensorElement({repr(sub_element)}, shape={shape}, "
                      f"symmetry={symmetry}{var_str})")

    @property
    def pullback(self):
        """Get pull back."""
        if len(self._symmetry) > 0:
            sub_element_value_shape = self.sub_elements[0].reference_value_shape
            for e in self.sub_elements:
                if e.reference_value_shape != sub_element_value_shape:
                    raise ValueError("Sub-elements must all have the same value size")
            symmetry = {}
            n = 0
            for i in np.ndindex(self._shape):
                if i in self._symmetry and self._symmetry[i] in symmetry:
                    symmetry[i] = symmetry[self._symmetry[i]]
                else:
                    symmetry[i] = n
                    n += 1
            return SymmetricPullback(self, symmetry)

        return self._sub_element.pullback

    def __repr__(self):
        """Doc."""
        return self._repr

    def variant(self):
        """Return the variant used to initialise the element."""
        return self._sub_element.variant()

    def mapping(self):
        """Doc."""
        return self._mapping

    def flattened_sub_element_mapping(self):
        """Doc."""
        return self._flattened_sub_element_mapping

    def extract_subelement_component(self, i):
        """Extract direct subelement index and subelement relative.

        component index for a given component index.
        """
        if isinstance(i, int):
            i = (i,)
        self._check_component(i)

        i = self.symmetry().get(i, i)
        l = len(self._shape)  # noqa: E741
        ii = i[:l]
        jj = i[l:]
        if ii not in self._sub_element_mapping:
            raise ValueError(f"Illegal component index {i}.")
        k = self._sub_element_mapping[ii]
        return (k, jj)

    def symmetry(self):
        r"""Return the symmetry dict, which is a mapping :math:`c_0 \\to c_1`.

        meaning that component :math:`c_0` is represented by component
        :math:`c_1`.
        A component is a tuple of one or more ints.
        """
        return self._symmetry

    def reconstruct(self, **kwargs):
        """Doc."""
        sub_element = self._sub_element.reconstruct(**kwargs)
        return TensorElement(sub_element, shape=self._shape, symmetry=self._symmetry)

    def __str__(self):
        """Format as string for pretty printing."""
        if self._symmetry:
            tmp = ", ".join("%s -> %s" % (a, b) for (a, b) in self._symmetry.items())
            sym = " with symmetries (%s)" % tmp
        else:
            sym = ""
        return ("<tensor element with shape %s of %s%s>" %
                (self.reference_value_shape, self._sub_element, sym))

    def shortstr(self):
        """Format as string for pretty printing."""
        if self._symmetry:
            tmp = ", ".join("%s -> %s" % (a, b) for (a, b) in self._symmetry.items())
            sym = " with symmetries (%s)" % tmp
        else:
            sym = ""
        return "Tensor<%s x %s%s>" % (self.reference_value_shape,
                                      self._sub_element.shortstr(), sym)



# Copyright (C) 2024 Pablo D. Brubeck
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2024

from FIAT.functional import (PointEvaluation, PointDerivative,
                             IntegralMoment,
                             IntegralMomentOfNormalDerivative)
from FIAT import finite_element, dual_set, macro, polynomial_set
from FIAT.reference_element import TRIANGLE, ufc_simplex
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
from FIAT.jacobi import eval_jacobi, eval_jacobi_batch, eval_jacobi_deriv_batch


class HCTDualSet(dual_set.DualSet):
    def __init__(self, ref_complex, degree, reduced=False):
        if reduced and degree != 3:
            raise ValueError("Reduced HCT only defined for degree = 3")
        if degree < 3:
            raise ValueError("HCT only defined for degree >= 3")
        ref_el = ref_complex.get_parent()
        if ref_el.get_shape() != TRIANGLE:
            raise ValueError("HCT only defined on triangles")
        top = ref_el.get_topology()
        verts = ref_el.get_vertices()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}

        # get first order jet at each vertex
        alphas = polynomial_set.mis(sd, 1)
        nodes = []
        for v in sorted(top[0]):
            pt = verts[v]
            cur = len(nodes)
            nodes.append(PointEvaluation(ref_el, pt))
            nodes.extend(PointDerivative(ref_el, pt, alpha) for alpha in alphas)
            entity_ids[0][v].extend(range(cur, len(nodes)))

        k = 2 if reduced else degree - 3
        facet = ufc_simplex(1)
        Q = create_quadrature(facet, degree-1+k)
        qpts = Q.get_points()
        xref = 2.0 * qpts - 1.0
        if reduced:
            f_at_qpts = eval_jacobi(0, 0, k, xref[:, 0])
            for e in sorted(top[1]):
                cur = len(nodes)
                nodes.append(IntegralMomentOfNormalDerivative(ref_el, e, Q, f_at_qpts))
                entity_ids[1][e].extend(range(cur, len(nodes)))
        else:
            phis = eval_jacobi_batch(1, 1, k, xref)
            dphis = eval_jacobi_deriv_batch(1, 1, k, xref)
            for e in sorted(top[1]):
                Q_mapped = FacetQuadratureRule(ref_el, 1, e, Q)
                scale = 2 / Q_mapped.jacobian_determinant()
                cur = len(nodes)
                nodes.extend(IntegralMomentOfNormalDerivative(ref_el, e, Q, phi) for phi in phis)
                nodes.extend(IntegralMoment(ref_el, Q_mapped, dphi * scale) for dphi in dphis[1:])
                entity_ids[1][e].extend(range(cur, len(nodes)))

            q = degree - 4
            if q >= 0:
                Q = create_quadrature(ref_complex, degree + q)
                Pq = polynomial_set.ONPolynomialSet(ref_el, q, scale=1)
                phis = Pq.tabulate(Q.get_points())[(0,) * sd]
                scale = 1 / ref_el.volume()
                cur = len(nodes)
                nodes.extend(IntegralMoment(ref_el, Q, phi * scale) for phi in phis)
                entity_ids[sd][0] = list(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class HsiehCloughTocher(finite_element.CiarletElement):
    """The HCT macroelement. For degree higher than 3, we implement the
    super-smooth C^1 space from Groselj and Knez (2022) on a barycentric split,
    although there the basis functions are positive on an incenter split.
    """
    def __init__(self, ref_el, degree=3, reduced=False):
        ref_complex = macro.AlfeldSplit(ref_el)
        dual = HCTDualSet(ref_complex, degree, reduced=reduced)
        poly_set = macro.CkPolynomialSet(ref_complex, degree, order=1, vorder=degree-1, variant="bubble")
        formdegree = 0
        super().__init__(poly_set, dual, degree, formdegree=formdegree)



from itertools import chain, product
from operator import methodcaller

import numpy

import FIAT
from FIAT.polynomial_set import mis
from FIAT.reference_element import TensorProductCell
from FIAT.orientation_utils import make_entity_permutations_tensorproduct

import gem
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase
from finat.point_set import PointSingleton, PointSet, TensorPointSet


class TensorProductElement(FiniteElementBase):

    def __init__(self, factors):
        super(TensorProductElement, self).__init__()
        self.factors = tuple(factors)

        shapes = [fe.value_shape for fe in self.factors if fe.value_shape != ()]
        if len(shapes) == 0:
            self._value_shape = ()
        elif len(shapes) == 1:
            self._value_shape = shapes[0]
        else:
            raise NotImplementedError("Only one nonscalar factor permitted!")

    @cached_property
    def cell(self):
        return TensorProductCell(*(fe.cell for fe in self.factors))

    @cached_property
    def complex(self):
        return TensorProductCell(*(fe.complex for fe in self.factors))

    @property
    def degree(self):
        return tuple(fe.degree for fe in self.factors)

    @cached_property
    def formdegree(self):
        if any(fe.formdegree is None for fe in self.factors):
            return None
        else:
            return sum(fe.formdegree for fe in self.factors)

    @cached_property
    def _entity_dofs(self):
        return productise(self.factors, methodcaller("entity_dofs"))

    @cached_property
    def _entity_support_dofs(self):
        return productise(self.factors, methodcaller("entity_support_dofs"))

    def entity_dofs(self):
        return self._entity_dofs

    @cached_property
    def entity_permutations(self):
        return compose_permutations(self.factors)

    def space_dimension(self):
        return numpy.prod([fe.space_dimension() for fe in self.factors])

    @property
    def index_shape(self):
        return tuple(chain.from_iterable(fe.index_shape for fe in self.factors))

    @property
    def value_shape(self):
        return self._value_shape

    @cached_property
    def fiat_equivalent(self):
        # FIAT TensorProductElement support only 2 factors
        A, B = self.factors
        return FIAT.TensorProductElement(A.fiat_equivalent, B.fiat_equivalent)

    def _factor_entity(self, entity):
        # Default entity
        if entity is None:
            entity = (self.cell.get_dimension(), 0)
        entity_dim, entity_id = entity

        # Factor entity
        assert isinstance(entity_dim, tuple)
        assert len(entity_dim) == len(self.factors)

        shape = tuple(len(c.get_topology()[d])
                      for c, d in zip(self.cell.cells, entity_dim))
        entities = list(zip(entity_dim, numpy.unravel_index(entity_id, shape)))
        return entities

    def _merge_evaluations(self, factor_results):
        # Spatial dimension
        dimension = self.cell.get_spatial_dimension()

        # Derivative order
        order = max(map(sum, chain(*factor_results)))

        # A list of slices that are used to select dimensions
        # corresponding to each subelement.
        dim_slices = TensorProductCell._split_slices([c.get_spatial_dimension()
                                                      for c in self.cell.cells])

        # A list of multiindices, one multiindex per subelement, each
        # multiindex describing the shape of basis functions of the
        # subelement.
        alphas = [fe.get_indices() for fe in self.factors]

        # A list of multiindices, one multiindex per subelement, each
        # multiindex describing the value shape of the subelement.
        zetas = [fe.get_value_indices() for fe in self.factors]

        multiindex = tuple(chain(*alphas, *zetas))
        result = {}
        for derivative in range(order + 1):
            for Delta in mis(dimension, derivative):
                # Split the multiindex for the subelements
                deltas = [Delta[s] for s in dim_slices]
                # Multiply the values from the subelements
                # Turn basis shape to free indices, select the
                # right derivative entry.
                scalar = gem.Product(*(gem.Indexed(fr[delta], alpha + zeta)
                                       for fr, delta, alpha, zeta
                                       in zip(factor_results, deltas, alphas, zetas)))
                # Wrap up non-point indices into shape.
                result[Delta] = gem.ComponentTensor(scalar, multiindex)
        return result

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        entities = self._factor_entity(entity)
        entity_dim, _ = zip(*entities)

        ps_factors = factor_point_set(self.cell, entity_dim, ps)

        factor_results = [fe.basis_evaluation(order, ps_, e)
                          for fe, ps_, e in zip(self.factors, ps_factors, entities)]

        return self._merge_evaluations(factor_results)

    def point_evaluation(self, order, point, entity=None, coordinate_mapping=None):
        entities = self._factor_entity(entity)
        entity_dim, _ = zip(*entities)

        # Split point expression
        assert len(self.cell.cells) == len(entity_dim)
        point_dims = [cell.construct_subelement(dim).get_spatial_dimension()
                      for cell, dim in zip(self.cell.cells, entity_dim)]
        assert isinstance(point, gem.Node) and point.shape == (sum(point_dims),)
        slices = TensorProductCell._split_slices(point_dims)
        point_factors = []
        for s in slices:
            point_factors.append(gem.ListTensor(
                [gem.Indexed(point, (i,))
                 for i in range(s.start, s.stop)]
            ))

        # Subelement results
        factor_results = [fe.point_evaluation(order, p_, e)
                          for fe, p_, e in zip(self.factors, point_factors, entities)]

        return self._merge_evaluations(factor_results)

    @property
    def dual_basis(self):
        # Outer product the dual bases of the factors
        qs, pss = zip(*(factor.dual_basis for factor in self.factors))
        ps = TensorPointSet(pss)
        # Naming as _merge_evaluations above
        alphas = [factor.get_indices() for factor in self.factors]
        zetas = [factor.get_value_indices() for factor in self.factors]
        Q = gem.ComponentTensor(
            # Index the factors by basis function and component so that we can reshape
            # into index-shape followed by value-shape
            gem.Product(*(q[alpha + zeta] for q, alpha, zeta in zip(qs, alphas, zetas))),
            tuple(chain(*alphas, *zetas))
        )
        return Q, ps

    @cached_property
    def mapping(self):
        mappings = [fe.mapping for fe in self.factors if fe.mapping != "affine"]
        if len(mappings) == 0:
            return "affine"
        elif len(mappings) == 1:
            return mappings[0]
        else:
            return None


def productise(factors, method):
    '''Tensor product the dict mapping topological entities to dofs across factors.

    :arg factors: element factors.
    :arg method: instance method to call on each factor to get dofs.'''
    shape = tuple(fe.space_dimension() for fe in factors)
    dofs = {}
    for dim in product(*[fe.cell.get_topology().keys()
                         for fe in factors]):
        dim_dofs = []
        topds = [method(fe)[d]
                 for fe, d in zip(factors, dim)]
        for tuple_ei in product(*[sorted(topd) for topd in topds]):
            tuple_vs = list(product(*[topd[ei]
                                      for topd, ei in zip(topds, tuple_ei)]))
            if tuple_vs:
                vs = list(numpy.ravel_multi_index(numpy.transpose(tuple_vs), shape))
                dim_dofs.append((tuple_ei, vs))
            else:
                dim_dofs.append((tuple_ei, []))
        # flatten entity numbers
        dofs[dim] = dict(enumerate(v for k, v in sorted(dim_dofs)))
    return dofs


def compose_permutations(factors):
    r"""For the :class:`TensorProductElement` object composed of factors,
    construct, for each dimension tuple, for each entity, and for each possible
    entity orientation combination, the DoF permutation list.

    :arg factors: element factors.
    :returns: entity_permutation dict of the :class:`TensorProductElement` object
        composed of factors.

    For tensor-product elements, one needs to consider two kinds of orientations:
    extrinsic orientations and intrinsic ("material") orientations.

    Example
    -------

    UFCQuadrilateral := UFCInterval x UFCInterval

    eo (extrinsic orientation): swap axes (X -> y, Y-> x)
    io (intrinsic orientation): reflect component intervals
    o (total orientation)     : (2 ** dim) * eo + io

    eo\\io    0      1      2      3

           1---3  0---2  3---1  2---0
      0    |   |  |   |  |   |  |   |
           0---2  1---3  2---0  3---1

           2---3  3---2  0---1  1---0
      1    |   |  |   |  |   |  |   |
           0---1  1---0  2---3  3---2

    .. code-block:: python3

        import FIAT
        import finat

        cell = FIAT.ufc_cell("interval")
        elem = finat.DiscontinuousLagrange(cell, 1)
        elem = finat.TensorProductElement([elem, elem])
        print(elem.entity_permutations)

    prints:

    {(0, 0): {0: {(0, 0, 0): []},
              1: {(0, 0, 0): []},
              2: {(0, 0, 0): []},
              3: {(0, 0, 0): []}},
     (0, 1): {0: {(0, 0, 0): [],
                  (0, 0, 1): []},
              1: {(0, 0, 0): [],
                  (0, 0, 1): []}},
     (1, 0): {0: {(0, 0, 0): [],
                  (0, 1, 0): []},
              1: {(0, 0, 0): [],
                  (0, 1, 0): []}},
     (1, 1): {0: {(0, 0, 0): [0, 1, 2, 3],
                  (0, 0, 1): [1, 0, 3, 2],
                  (0, 1, 0): [2, 3, 0, 1],
                  (0, 1, 1): [3, 2, 1, 0],
                  (1, 0, 0): [0, 2, 1, 3],
                  (1, 0, 1): [2, 0, 3, 1],
                  (1, 1, 0): [1, 3, 0, 2],
                  (1, 1, 1): [3, 1, 2, 0]}}}

    """
    permutations = {}
    cells = [fe.cell for fe in factors]
    for dim in product(*[cell.get_topology().keys() for cell in cells]):
        dim_permutations = []
        e_o_p_maps = [fe.entity_permutations[d] for fe, d in zip(factors, dim)]
        for e_tuple in product(*[sorted(e_o_p_map) for e_o_p_map in e_o_p_maps]):
            o_p_maps = [e_o_p_map[e] for e_o_p_map, e in zip(e_o_p_maps, e_tuple)]
            o_tuple_perm_map = make_entity_permutations_tensorproduct(cells, dim, o_p_maps)
            dim_permutations.append((e_tuple, o_tuple_perm_map))
        permutations[dim] = dict(enumerate(v for k, v in sorted(dim_permutations)))
    return permutations


def factor_point_set(product_cell, product_dim, point_set):
    """Factors a point set for the product element into a point sets for
    each subelement.

    :arg product_cell: a TensorProductCell
    :arg product_dim: entity dimension for the product cell
    :arg point_set: point set for the product element
    """
    assert len(product_cell.cells) == len(product_dim)
    point_dims = [cell.construct_subelement(dim).get_spatial_dimension()
                  for cell, dim in zip(product_cell.cells, product_dim)]

    if isinstance(point_set, TensorPointSet) and \
       len(product_cell.cells) == len(point_set.factors):
        # Just give the factors asserting matching dimensions.
        assert len(point_set.factors) == len(point_dims)
        assert all(ps.dimension == dim
                   for ps, dim in zip(point_set.factors, point_dims))
        return point_set.factors

    # Split the point coordinates along the point dimensions
    # required by the subelements.
    assert point_set.dimension == sum(point_dims)
    slices = TensorProductCell._split_slices(point_dims)
    if isinstance(point_set, PointSingleton):
        return [PointSingleton(point_set.point[s]) for s in slices]
    elif isinstance(point_set, (PointSet, TensorPointSet)):
        # Use the same point index for the new point sets.
        result = []
        for s in slices:
            ps = PointSet(point_set.points[:, s])
            ps.indices = point_set.indices
            result.append(ps)
        return result

    raise NotImplementedError("How to tabulate TensorProductElement on %s?" % (type(point_set).__name__,))



import numpy as np
from FIAT import ufc_simplex, ArnoldWinther, create_quadrature, expansions


def test_dofs():
    line = ufc_simplex(1)
    T = ufc_simplex(2)
    T.vertices = ((0.0, 0.0), (1.0, 0.25), (-0.75, 1.1))
    AW = ArnoldWinther(T, 3)

    # check Kronecker property at vertices

    bases = [[[1, 0], [0, 0]], [[0, 1], [1, 0]], [[0, 0], [0, 1]]]

    vert_vals = AW.tabulate(0, T.vertices)[(0, 0)]
    for i in range(3):
        for j in range(3):
            assert np.allclose(vert_vals[3*i+j, :, :, i], bases[j])
            for k in (1, 2):
                assert np.allclose(vert_vals[3*i+j, :, :, (i+k) % 3], np.zeros((2, 2)))

    # check edge moments
    Qline = create_quadrature(line, 6)

    linebfs = expansions.LineExpansionSet(line)
    linevals = linebfs.tabulate(1, Qline.pts)

    # n, n moments
    for ed in range(3):
        n = T.compute_scaled_normal(ed)
        wts = np.asarray(Qline.wts)
        nqpline = len(wts)

        vals = AW.tabulate(0, Qline.pts, (1, ed))[(0, 0)]
        nnvals = np.zeros((30, nqpline))
        for i in range(30):
            for j in range(len(wts)):
                nnvals[i, j] = n @ vals[i, :, :, j] @ n

        nnmoments = np.zeros((30, 2))

        for bf in range(30):
            for k in range(nqpline):
                for m in (0, 1):
                    nnmoments[bf, m] += wts[k] * nnvals[bf, k] * linevals[m, k]

        for bf in range(30):
            if bf != AW.dual.entity_ids[1][ed][0] and bf != AW.dual.entity_ids[1][ed][2]:
                assert np.allclose(nnmoments[bf, :], np.zeros(2))

    # n, t moments
    for ed in range(3):
        n = T.compute_scaled_normal(ed)
        t = T.compute_edge_tangent(ed)
        wts = np.asarray(Qline.wts)
        nqpline = len(wts)

        vals = AW.tabulate(0, Qline.pts, (1, ed))[(0, 0)]
        ntvals = np.zeros((30, nqpline))
        for i in range(30):
            for j in range(len(wts)):
                ntvals[i, j] = n @ vals[i, :, :, j] @ t

        ntmoments = np.zeros((30, 2))

        for bf in range(30):
            for k in range(nqpline):
                for m in (0, 1):
                    ntmoments[bf, m] += wts[k] * ntvals[bf, k] * linevals[m, k]

        for bf in range(30):
            if bf != AW.dual.entity_ids[1][ed][1] and bf != AW.dual.entity_ids[1][ed][3]:
                assert np.allclose(ntmoments[bf, :], np.zeros(2))

    # check internal dofs
    ns = list(map(T.compute_scaled_normal, range(3)))
    Q = create_quadrature(T, 3)
    qpvals = AW.tabulate(0, Q.pts)[(0, 0)]
    const_moms = qpvals @ Q.wts / T.volume()
    nn_moms = const_moms.copy()
    for j in range(2):
        for i in range(2):
            comp = np.outer(ns[i+1], ns[j+1])
            nn_moms[:, i, j] = np.tensordot(const_moms, comp, ((1, 2), (0, 1)))
    assert np.allclose(nn_moms[:21], np.zeros((21, 2, 2)))
    assert np.allclose(nn_moms[24:], np.zeros((6, 2, 2)))
    assert np.allclose(nn_moms[21:24, 0, 0], np.asarray([1, 0, 0]))
    assert np.allclose(nn_moms[21:24, 0, 1], np.asarray([0, 1, 0]))
    assert np.allclose(nn_moms[21:24, 1, 0], np.asarray([0, 1, 0]))
    assert np.allclose(nn_moms[21:24, 1, 1], np.asarray([0, 0, 1]))


def frob(a, b):
    return a.ravel() @ b.ravel()


def test_projection():
    T = ufc_simplex(2)
    T.vertices = ((0.0, 0.0), (1.0, 0.0), (0.5, 2.1))

    AW = ArnoldWinther(T, 3)

    Q = create_quadrature(T, 6)
    qpts = np.asarray(Q.pts)
    qwts = np.asarray(Q.wts)
    nqp = len(Q.wts)

    nbf = 24
    m = np.zeros((nbf, nbf))
    b = np.zeros((24,))
    rhs_vals = np.zeros((2, 2, nqp))

    bfvals = AW.tabulate(0, qpts)[(0, 0)][:nbf, :, :, :]

    for i in range(nbf):
        for j in range(nbf):
            for k in range(nqp):
                m[i, j] += qwts[k] * frob(bfvals[i, :, :, k],
                                          bfvals[j, :, :, k])

    assert np.linalg.cond(m) < 1.e12

    comps = [(0, 0), (0, 1), (0, 0)]

    # loop over monomials up to degree 2
    for deg in range(3):
        for jj in range(deg+1):
            ii = deg-jj
            for comp in comps:
                b[:] = 0.0
                # set RHS (symmetrically) to be the monomial in
                # the proper component.
                rhs_vals[comp] = qpts[:, 0]**ii * qpts[:, 1]**jj
                rhs_vals[tuple(reversed(comp))] = rhs_vals[comp]
                for i in range(nbf):
                    for k in range(nqp):
                        b[i] += qwts[k] * frob(bfvals[i, :, :, k],
                                               rhs_vals[:, :, k])
                x = np.linalg.solve(m, b)

                sol_at_qpts = np.zeros(rhs_vals.shape)
                for i in range(nbf):
                    for k in range(nqp):
                        sol_at_qpts[:, :, k] += x[i] * bfvals[i, :, :, k]

                diff = sol_at_qpts - rhs_vals
                err = 0.0
                for k in range(nqp):
                    err += qwts[k] * frob(diff[:, :, k], diff[:, :, k])

                assert np.sqrt(err) < 1.e-12



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT import (polynomial_set, expansions, dual_set,
                  finite_element, functional, macro)
import numpy
from itertools import chain
from FIAT.check_format_variant import check_format_variant
from FIAT.quadrature_schemes import create_quadrature
from FIAT.quadrature import FacetQuadratureRule


def NedelecSpace2D(ref_el, degree):
    """Constructs a basis for the 2d H(curl) space of the first kind
    which is (P_{degree-1})^2 + P_{degree-1} rot( x )"""
    sd = ref_el.get_spatial_dimension()
    if sd != 2:
        raise Exception("NedelecSpace2D requires 2d reference element")

    k = degree - 1
    vec_Pkp1 = polynomial_set.ONPolynomialSet(ref_el, k + 1, (sd,))

    dimPkp1 = expansions.polynomial_dimension(ref_el, k + 1)
    dimPk = expansions.polynomial_dimension(ref_el, k)
    dimPkm1 = expansions.polynomial_dimension(ref_el, k - 1)

    vec_Pk_indices = list(chain(*(range(i * dimPkp1, i * dimPkp1 + dimPk)
                                  for i in range(sd))))
    vec_Pk_from_Pkp1 = vec_Pkp1.take(vec_Pk_indices)

    Pkp1 = polynomial_set.ONPolynomialSet(ref_el, k + 1)
    PkH = Pkp1.take(list(range(dimPkm1, dimPk)))

    Q = create_quadrature(ref_el, 2 * (k + 1))
    Qpts, Qwts = Q.get_points(), Q.get_weights()

    PkH_at_Qpts = PkH.tabulate(Qpts)[(0,) * sd]
    Pkp1_at_Qpts = Pkp1.tabulate(Qpts)[(0,) * sd]

    CrossX = numpy.dot(numpy.array([[0.0, 1.0], [-1.0, 0.0]]), Qpts.T)
    PkHCrossX_at_Qpts = PkH_at_Qpts[:, None, :] * CrossX[None, :, :]
    PkHCrossX_coeffs = numpy.dot(numpy.multiply(PkHCrossX_at_Qpts, Qwts),
                                 Pkp1_at_Qpts.T)
    PkHcrossX = polynomial_set.PolynomialSet(ref_el,
                                             k + 1,
                                             k + 1,
                                             vec_Pkp1.get_expansion_set(),
                                             PkHCrossX_coeffs)

    return polynomial_set.polynomial_set_union_normalized(vec_Pk_from_Pkp1,
                                                          PkHcrossX)


def NedelecSpace3D(ref_el, degree):
    """Constructs a nodal basis for the 3d first-kind Nedelec space"""
    sd = ref_el.get_spatial_dimension()
    if sd != 3:
        raise Exception("NedelecSpace3D requires 3d reference element")

    k = degree - 1
    vec_Pkp1 = polynomial_set.ONPolynomialSet(ref_el, k + 1, (sd,))

    dimPkp1 = expansions.polynomial_dimension(ref_el, k + 1)
    dimPk = expansions.polynomial_dimension(ref_el, k)
    dimPkm1 = expansions.polynomial_dimension(ref_el, k - 1)

    vec_Pk_indices = list(chain(*(range(i * dimPkp1, i * dimPkp1 + dimPk)
                                  for i in range(sd))))
    vec_Pk = vec_Pkp1.take(vec_Pk_indices)

    vec_Pke_indices = list(chain(*(range(i * dimPkp1 + dimPkm1, i * dimPkp1 + dimPk)
                                   for i in range(sd))))
    vec_Pke = vec_Pkp1.take(vec_Pke_indices)

    Pkp1 = polynomial_set.ONPolynomialSet(ref_el, k + 1)

    Q = create_quadrature(ref_el, 2 * (k + 1))
    Qpts, Qwts = Q.get_points(), Q.get_weights()

    Pke_qpts = vec_Pke.tabulate(Qpts)[(0,) * sd]
    Pkp1_at_Qpts = Pkp1.tabulate(Qpts)[(0,) * sd]

    x = Qpts.T
    PkCrossX_at_Qpts = numpy.cross(Pke_qpts, x[None, :, :], axis=1)
    PkCrossXcoeffs = numpy.dot(numpy.multiply(PkCrossX_at_Qpts, Qwts), Pkp1_at_Qpts.T)

    PkCrossX = polynomial_set.PolynomialSet(ref_el,
                                            k + 1,
                                            k + 1,
                                            vec_Pkp1.get_expansion_set(),
                                            PkCrossXcoeffs)
    return polynomial_set.polynomial_set_union_normalized(vec_Pk, PkCrossX)


class NedelecDual(dual_set.DualSet):
    """Dual basis for first-kind Nedelec."""

    def __init__(self, ref_el, degree, variant, interpolant_deg):
        nodes = []
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()

        entity_ids = {}
        # set to empty
        for dim in top:
            entity_ids[dim] = {}
            for entity in top[dim]:
                entity_ids[dim][entity] = []

        if variant == "integral":
            # edge nodes are \int_F v\cdot t p ds where p \in P_{q-1}(edge)
            # degree is q - 1

            # face nodes are \int_F v\cdot p dA where p \in P_{q-2}(f)^3 with p \cdot n = 0 (cmp. Monk)
            # these are equivalent to dofs from Fenics book defined by
            # \int_F v\times n \cdot p ds where p \in P_{q-2}(f)^2
            for dim in range(1, sd):
                phi_deg = degree - dim
                if phi_deg >= 0:
                    facet = ref_el.construct_subelement(dim)
                    Q_ref = create_quadrature(facet, interpolant_deg + phi_deg)
                    Pqmd = polynomial_set.ONPolynomialSet(facet, phi_deg, (dim,))
                    Phis = Pqmd.tabulate(Q_ref.get_points())[(0,) * dim]
                    Phis = numpy.transpose(Phis, (0, 2, 1))

                    for entity in top[dim]:
                        cur = len(nodes)
                        Q = FacetQuadratureRule(ref_el, dim, entity, Q_ref)
                        Jdet = Q.jacobian_determinant()
                        R = numpy.array(ref_el.compute_tangents(dim, entity))
                        phis = numpy.dot(Phis, R / Jdet)
                        phis = numpy.transpose(phis, (0, 2, 1))
                        nodes.extend(functional.FrobeniusIntegralMoment(ref_el, Q, phi)
                                     for phi in phis)
                        entity_ids[dim][entity] = list(range(cur, len(nodes)))

        elif variant == "point":
            for i in top[1]:
                cur = len(nodes)
                # points to specify P_k on each edge
                pts_cur = ref_el.make_points(1, i, degree + 1)
                nodes.extend(functional.PointEdgeTangentEvaluation(ref_el, i, pt)
                             for pt in pts_cur)
                entity_ids[1][i] = list(range(cur, len(nodes)))

            if sd > 2 and degree > 1:  # face tangents
                for i in top[2]:  # loop over faces
                    cur = len(nodes)
                    pts_cur = ref_el.make_points(2, i, degree + 1)
                    nodes.extend(functional.PointFaceTangentEvaluation(ref_el, i, k, pt)
                                 for k in range(2)  # loop over tangents
                                 for pt in pts_cur  # loop over points
                                 )
                    entity_ids[2][i] = list(range(cur, len(nodes)))

        # internal nodes. These are \int_T v \cdot p dx where p \in P_{q-d}^3(T)
        phi_deg = degree - sd
        if phi_deg >= 0:
            if interpolant_deg is None:
                interpolant_deg = degree

            cell = ref_el.construct_subelement(sd)
            Q_ref = create_quadrature(cell, interpolant_deg + phi_deg)
            Pqmd = polynomial_set.ONPolynomialSet(cell, phi_deg)
            Phis = Pqmd.tabulate(Q_ref.get_points())[(0,) * sd]

            for entity in top[sd]:
                Q = FacetQuadratureRule(ref_el, sd, entity, Q_ref)
                cur = len(nodes)
                nodes.extend(functional.IntegralMoment(ref_el, Q, phi, (d,), (sd,))
                             for d in range(sd)
                             for phi in Phis)
                entity_ids[sd][entity] = list(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class Nedelec(finite_element.CiarletElement):
    """
    Nedelec finite element

    :arg ref_el: The reference element.
    :arg degree: The degree.
    :arg variant: optional variant specifying the types of nodes.

    variant can be chosen from ["point", "integral", "integral(q)"]
    "point" -> dofs are evaluated by point evaluation. Note that this variant
    has suboptimal convergence order in the H(curl)-norm
    "integral" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence.
    "integral(q)" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence plus q. You might want to choose a high
    quadrature degree to make sure that expressions will be interpolated
    exactly. This is important when you want to have (nearly) curl-preserving
    interpolation.
    """

    def __init__(self, ref_el, degree, variant=None):
        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        if ref_el.is_macrocell():
            base_element = Nedelec(ref_el.get_parent(), degree)
            poly_set = macro.MacroPolynomialSet(ref_el, base_element)
        elif ref_el.get_spatial_dimension() == 3:
            poly_set = NedelecSpace3D(ref_el, degree)
        elif ref_el.get_spatial_dimension() == 2:
            poly_set = NedelecSpace2D(ref_el, degree)
        else:
            raise Exception("Not implemented")
        dual = NedelecDual(ref_el, degree, variant, interpolant_deg)
        formdegree = 1  # 1-form
        super().__init__(poly_set, dual, degree, formdegree, mapping="covariant piola")



"""This module defines the UFL finite element classes."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from abc import abstractmethod, abstractproperty
from hashlib import md5

from ufl import pullback
from ufl.cell import AbstractCell, as_cell
from ufl.finiteelement import AbstractFiniteElement
from ufl.utils.sequences import product


class FiniteElementBase(AbstractFiniteElement):
    """Base class for all finite elements."""
    __slots__ = ("_family", "_cell", "_degree", "_quad_scheme",
                 "_reference_value_shape")

    # TODO: Not all these should be in the base class! In particular
    # family, degree, and quad_scheme do not belong here.
    def __init__(self, family, cell, degree, quad_scheme,
                 reference_value_shape):
        """Initialize basic finite element data."""
        if not (degree is None or isinstance(degree, (int, tuple))):
            raise ValueError("Invalid degree type.")
        if not isinstance(reference_value_shape, tuple):
            raise ValueError("Invalid reference_value_shape type.")

        if cell is not None:
            cell = as_cell(cell)
            if not isinstance(cell, AbstractCell):
                raise ValueError("Invalid cell type.")

        self._family = family
        self._cell = cell
        self._degree = degree
        self._reference_value_shape = reference_value_shape
        self._quad_scheme = quad_scheme

    @abstractmethod
    def __repr__(self):
        """Format as string for evaluation as Python object."""
        pass

    @abstractproperty
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        pass

    @abstractmethod
    def mapping(self):
        """Return the mapping type for this element."""
        pass

    def _is_globally_constant(self):
        """Check if the element is a global constant.

        For Real elements, this should return True.
        """
        return False

    def _is_linear(self):
        """Check if the element is Lagrange degree 1."""
        return False

    def _ufl_hash_data_(self):
        """Doc."""
        return repr(self)

    def _ufl_signature_data_(self):
        """Doc."""
        return repr(self)

    def __hash__(self):
        """Compute hash code for insertion in hashmaps."""
        return int.from_bytes(md5(self._ufl_hash_data_().encode()).digest(), byteorder='big')

    def __eq__(self, other):
        """Compute element equality for insertion in hashmaps."""
        return type(self) is type(other) and self._ufl_hash_data_() == other._ufl_hash_data_()

    def __ne__(self, other):
        """Compute element inequality for insertion in hashmaps."""
        return not self.__eq__(other)

    def __lt__(self, other):
        """Compare elements by repr, to give a natural stable sorting."""
        return repr(self) < repr(other)

    def family(self):  # FIXME: Undefined for base?
        """Return finite element family."""
        return self._family

    def variant(self):
        """Return the variant used to initialise the element."""
        return None

    def degree(self, component=None):
        """Return polynomial degree of finite element."""
        return self._degree

    def quadrature_scheme(self):
        """Return quadrature scheme of finite element."""
        return self._quad_scheme

    @property
    def cell(self):
        """Return cell of finite element."""
        return self._cell

    def is_cellwise_constant(self, component=None):
        """Return whether the basis functions of this element is spatially constant over each cell."""
        return self._is_globally_constant() or self.degree() == 0

    @property
    def reference_value_shape(self):
        """Return the shape of the value space on the reference cell."""
        return self._reference_value_shape

    @property
    def reference_value_size(self):
        """Return the integer product of the reference value shape."""
        return product(self.reference_value_shape)

    def symmetry(self):  # FIXME: different approach
        r"""Return the symmetry dict.

        This is a mapping :math:`c_0 \\to c_1`
        meaning that component :math:`c_0` is represented by component
        :math:`c_1`.
        A component is a tuple of one or more ints.
        """
        return {}

    def _check_component(self, domain, i):
        """Check that component index i is valid."""
        sh = self.value_shape(domain.geometric_dimension())
        r = len(sh)
        if not (len(i) == r and all(j < k for (j, k) in zip(i, sh))):
            raise ValueError(
                f"Illegal component index {i} (value rank {len(i)}) "
                f"for element (value rank {r}).")

    def extract_subelement_component(self, domain, i):
        """Extract direct subelement index and subelement relative component index for a given component index."""
        if isinstance(i, int):
            i = (i,)
        self._check_component(domain, i)
        return (None, i)

    def extract_component(self, domain, i):
        """Recursively extract component index relative to a (simple) element.

        and that element for given value component index.
        """
        if isinstance(i, int):
            i = (i,)
        self._check_component(domain, i)
        return (i, self)

    def _check_reference_component(self, i):
        """Check that reference component index i is valid."""
        sh = self.reference_value_shape
        r = len(sh)
        if not (len(i) == r and all(j < k for (j, k) in zip(i, sh))):
            raise ValueError(
                f"Illegal component index {i} (value rank {len(i)}) "
                f"for element (value rank {r}).")

    def extract_subelement_reference_component(self, i):
        """Extract direct subelement index and subelement relative.

        reference component index for a given reference component index.
        """
        if isinstance(i, int):
            i = (i,)
        self._check_reference_component(i)
        return (None, i)

    def extract_reference_component(self, i):
        """Recursively extract reference component index relative to a (simple) element.

        and that element for given reference value component index.
        """
        if isinstance(i, int):
            i = (i,)
        self._check_reference_component(i)
        return (i, self)

    @property
    def num_sub_elements(self):
        """Return number of sub-elements."""
        return 0

    @property
    def sub_elements(self):
        """Return list of sub-elements."""
        return []

    def __add__(self, other):
        """Add two elements, creating an enriched element."""
        if not isinstance(other, FiniteElementBase):
            raise ValueError(f"Can't add element and {other.__class__}.")
        from finat.ufl import EnrichedElement
        return EnrichedElement(self, other)

    def __mul__(self, other):
        """Multiply two elements, creating a mixed element."""
        if not isinstance(other, FiniteElementBase):
            raise ValueError("Can't multiply element and {other.__class__}.")
        from finat.ufl import MixedElement
        return MixedElement(self, other)

    def __getitem__(self, index):
        """Restrict finite element to a subdomain, subcomponent or topology (cell)."""
        from finat.ufl.restrictedelement import valid_restriction_domains
        if index in valid_restriction_domains:
            from finat.ufl import RestrictedElement
            return RestrictedElement(self, index)
        else:
            raise KeyError(f"Invalid index for restriction: {repr(index)}")

    def __iter__(self):
        """Iter."""
        raise TypeError(f"'{type(self).__name__}' object is not iterable")

    @property
    def embedded_superdegree(self):
        """Doc."""
        return self.degree()

    @property
    def embedded_subdegree(self):
        """Doc."""
        return self.degree()

    @property
    def pullback(self):
        """Get the pull back."""
        if self.mapping() == "identity":
            return pullback.identity_pullback
        elif self.mapping() == "L2 Piola":
            return pullback.l2_piola
        elif self.mapping() == "covariant Piola":
            return pullback.covariant_piola
        elif self.mapping() == "contravariant Piola":
            return pullback.contravariant_piola
        elif self.mapping() == "double covariant Piola":
            return pullback.double_covariant_piola
        elif self.mapping() == "double contravariant Piola":
            return pullback.double_contravariant_piola
        elif self.mapping() == "covariant contravariant Piola":
            return pullback.covariant_contravariant_piola
        elif self.mapping() == "custom":
            return pullback.custom_pullback
        elif self.mapping() == "physical":
            return pullback.physical_pullback

        raise ValueError(f"Unsupported mapping: {self.mapping()}")



# Copyright (C) 2024 Pablo D. Brubeck
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2024

from FIAT import finite_element, dual_set, polynomial_set
from FIAT.functional import ComponentPointEvaluation, PointDivergence
from FIAT.quadrature_schemes import create_quadrature
from FIAT.macro import CkPolynomialSet, AlfeldSplit

import numpy


def AlfeldSorokinaSpace(ref_el, degree):
    """Return a vector-valued C0 PolynomialSet on an Alfeld split with C0
    divergence. This works on any simplex and for all polynomial degrees."""
    ref_complex = AlfeldSplit(ref_el)
    sd = ref_complex.get_spatial_dimension()
    C0 = CkPolynomialSet(ref_complex, degree, order=0, shape=(sd,), variant="bubble")
    expansion_set = C0.get_expansion_set()
    num_members = C0.get_num_members()
    coeffs = C0.get_coeffs()

    facet_el = ref_complex.construct_subelement(sd-1)
    phi = polynomial_set.ONPolynomialSet(facet_el, 0 if sd == 1 else degree-1)
    Q = create_quadrature(facet_el, 2 * phi.degree)
    qpts, qwts = Q.get_points(), Q.get_weights()
    phi_at_qpts = phi.tabulate(qpts)[(0,) * (sd-1)]
    weights = numpy.multiply(phi_at_qpts, qwts)

    rows = []
    for facet in ref_complex.get_interior_facets(sd-1):
        n = ref_complex.compute_normal(facet)
        jumps = expansion_set.tabulate_normal_jumps(degree, qpts, facet, order=1)
        div_jump = n[:, None, None] * jumps[1][None, ...]
        r = numpy.tensordot(div_jump, weights, axes=(-1, -1))
        rows.append(r.reshape(num_members, -1).T)

    if len(rows) > 0:
        dual_mat = numpy.vstack(rows)
        nsp = polynomial_set.spanning_basis(dual_mat, nullspace=True)
        coeffs = numpy.tensordot(nsp, coeffs, axes=(-1, 0))
    return polynomial_set.PolynomialSet(ref_complex, degree, degree, expansion_set, coeffs)


class AlfeldSorokinaDualSet(dual_set.DualSet):
    def __init__(self, ref_el, degree):
        if degree != 2:
            raise NotImplementedError(f"{type(self).__name__} only defined for degree = 2")

        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}

        nodes = []
        for dim in sorted(top):
            for entity in sorted(top[dim]):
                cur = len(nodes)

                dpts = ref_el.make_points(dim, entity, degree-1)
                nodes.extend(PointDivergence(ref_el, pt) for pt in dpts)

                pts = ref_el.make_points(dim, entity, degree)
                nodes.extend(ComponentPointEvaluation(ref_el, k, (sd,), pt)
                             for pt in pts for k in range(sd))
                entity_ids[dim][entity].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class AlfeldSorokina(finite_element.CiarletElement):
    """The Alfeld-Sorokina C0 quadratic macroelement with C0 divergence.

    This element belongs to a Stokes complex, and is paired with CG1(Alfeld).
    """
    def __init__(self, ref_el, degree=2):
        dual = AlfeldSorokinaDualSet(ref_el, degree)
        poly_set = AlfeldSorokinaSpace(ref_el, degree)
        formdegree = ref_el.get_spatial_dimension() - 1  # (n-1)-form
        super().__init__(poly_set, dual, degree, formdegree,
                         mapping="contravariant piola")



from functools import singledispatch, reduce

import numpy
import sympy
try:
    import symengine
except ImportError:
    class Mock:
        def __getattribute__(self, name):
            return Mock
    symengine = Mock()

import gem


@singledispatch
def sympy2gem(node, self):
    raise AssertionError("sympy/symengine node expected, got %s" % type(node))


@sympy2gem.register(sympy.Expr)
@sympy2gem.register(symengine.Expr)
def sympy2gem_expr(node, self):
    raise NotImplementedError("no handler for sympy/symengine node type %s" % type(node))


@sympy2gem.register(sympy.Add)
@sympy2gem.register(symengine.Add)
def sympy2gem_add(node, self):
    return gem.Sum(*map(self, node.args))


@sympy2gem.register(sympy.Mul)
@sympy2gem.register(symengine.Mul)
def sympy2gem_mul(node, self):
    return gem.Product(*map(self, node.args))


@sympy2gem.register(sympy.Pow)
@sympy2gem.register(symengine.Pow)
def sympy2gem_pow(node, self):
    return gem.Power(*map(self, node.args))


@sympy2gem.register(sympy.logic.boolalg.BooleanTrue)
@sympy2gem.register(sympy.logic.boolalg.BooleanFalse)
@sympy2gem.register(bool)
def sympy2gem_boolean(node, self):
    return gem.Literal(bool(node))


@sympy2gem.register(sympy.Integer)
@sympy2gem.register(symengine.Integer)
@sympy2gem.register(int)
def sympy2gem_integer(node, self):
    return gem.Literal(int(node))


@sympy2gem.register(sympy.Float)
@sympy2gem.register(symengine.Float)
@sympy2gem.register(float)
def sympy2gem_float(node, self):
    return gem.Literal(float(node))


@sympy2gem.register(sympy.Symbol)
@sympy2gem.register(symengine.Symbol)
def sympy2gem_symbol(node, self):
    return self.bindings[node]


@sympy2gem.register(sympy.Rational)
@sympy2gem.register(symengine.Rational)
def sympy2gem_rational(node, self):
    return gem.Division(*(map(self, node.as_numer_denom())))


@sympy2gem.register(sympy.Abs)
@sympy2gem.register(symengine.Abs)
def sympy2gem_abs(node, self):
    return gem.MathFunction("abs", *map(self, node.args))


@sympy2gem.register(sympy.Not)
@sympy2gem.register(symengine.Not)
def sympy2gem_not(node, self):
    return gem.LogicalNot(*map(self, node.args))


@sympy2gem.register(sympy.Or)
@sympy2gem.register(symengine.Or)
def sympy2gem_or(node, self):
    return reduce(gem.LogicalOr, map(self, node.args))


@sympy2gem.register(sympy.And)
@sympy2gem.register(symengine.And)
def sympy2gem_and(node, self):
    return reduce(gem.LogicalAnd, map(self, node.args))


@sympy2gem.register(sympy.Eq)
@sympy2gem.register(symengine.Eq)
def sympy2gem_eq(node, self):
    return gem.Comparison("==", *map(self, node.args))


@sympy2gem.register(sympy.Gt)
def sympy2gem_gt(node, self):
    return gem.Comparison(">", *map(self, node.args))


@sympy2gem.register(sympy.Ge)
def sympy2gem_ge(node, self):
    return gem.Comparison(">=", *map(self, node.args))


@sympy2gem.register(sympy.Lt)
@sympy2gem.register(symengine.Lt)
def sympy2gem_lt(node, self):
    return gem.Comparison("<", *map(self, node.args))


@sympy2gem.register(sympy.Le)
@sympy2gem.register(symengine.Le)
def sympy2gem_le(node, self):
    return gem.Comparison("<=", *map(self, node.args))


@sympy2gem.register(sympy.Piecewise)
@sympy2gem.register(symengine.Piecewise)
def sympy2gem_conditional(node, self):
    expr = None
    pieces = []
    for v, c in node.args:
        if isinstance(c, (bool, numpy.bool, sympy.logic.boolalg.BooleanTrue)) and c:
            expr = self(v)
            break
        pieces.append((v, c))
    if expr is None:
        expr = gem.Literal(float("nan"))
    for v, c in reversed(pieces):
        expr = gem.Conditional(self(c), self(v), expr)
    return expr


@sympy2gem.register(sympy.ITE)
def sympy2gem_ifthenelse(node, self):
    return gem.Conditional(*map(self, node.args))
